#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
import esppy
from . import utils as tm


class TestCounterWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('counter_window.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto', n_threads='1')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',
                               schema=('ID*:int32','symbol:string','currency:int32',
                                       'udate:int64','msecs:int32','price:double','quant:int32',
                                       'venue:int32','broker:int32','buyer:int32','seller:int32',
                                       'buysellflg:int32'))

        counter_win = esp.CounterWindow(name='count',
                                        count_interval='5 seconds',
                                        clear_interval='20 seconds')

        compute_win = esp.ComputeWindow(name='compute_count',
                                        schema=('input*:string','totalCount:int32'))
        compute_win.add_field_expression('totalCount')
        compute_win.add_connector('fs',
                                   conn_name='sub', conn_type='subscribe',
                                   type='sub',
                                   fstype='csv',
                                   fsname='result.out',
                                   snapshot=True,
                                   collapse=True)

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(counter_win)
        cq.add_window(compute_win)

        src.add_target(counter_win)
        counter_win.add_target(compute_win)

        self._test_model_file('counter_window.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
import esppy
from . import utils as tm


class TestFunctionalWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('functional_window.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto', n_threads='1')

        cq = esp.ContinuousQuery(name='cq_01',trace='src_win transform')

        src = esp.SourceWindow(name='src_win',
                               schema=('id*:int64','symbol:string','quant:int32',
                                       'dateD:date','price:double','timeD:stamp'))
        src.add_connector('fs',
                          conn_name='pub', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='functional_window.csv',
                          dateformat='%Y-%m-%d %H:%M:%S')

        functional_win = esp.FunctionalWindow(name='transform',
                                              schema=('id*:int64', 'symbol:string',
                                                      'total:string', 'timeD:stamp',
                                                      'timeString:string',
                                                      'dateD:date', 'dateSstring:string'))
        functional_win.set_function_context_functions(
            total='string(product($quant, $price))',
            timeString='timeGmtString(quotient($timeD,1000000),\'%Y-%m-%d %H:%M:%S\')',
            dateSstring='timeGmtString($dateD,\'%Y-%m-%d %H:%M:%S\')')
        functional_win.add_connector('fs',
                                     conn_name='sub', conn_type='subscribe',
                                     type='sub',
                                     fstype='csv',
                                     fsname='result.out',
                                     snapshot=True,
                                     dateformat='%Y-%m-%d %H:%M:%S')

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(functional_win)

        src.add_target(functional_win)

        self._test_model_file('functional_window.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP PI Asset Framework Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class PISubscriber(Connector):
    '''
    Subscribe to operations from a PI Asset Framework (AF) server

    Parameters
    ----------
    afelement : string
        Specifies the AF element or element template name.
        Wildcards are supported.
    iselementtemplate : boolean
        Specifies whether the afelement parameter is an element template
        name. By default, the afelement parameter specifies an element name.
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    rmretdel : boolean, optional
        Removes all delete events from event blocks received by the
        subscriber that were introduced by a window retention policy.
    pisystem : string, optional
        Specifies the PI system. The default is the PI system that is
        configured in the PI AF client.
    afdatabase : string, optional
        Specifies the AF database. The default is the AF database that is
        configured in the PI AF client.
    afrootelement : string, optional
        Specifies the root element in the AF hierarchy from which to search
        for parameter afelement. The default is the top-level element
        in the AF database.
    afattribute : string, optional
        Specifies a specific attribute in the element. The default is
        all attributes in the element.
    configfilesection : string, optional
        Specifies the name of the section in the config file to parse for
        configuration parameters. Specify the value as [configfilesection].

    Returns
    -------
    :class:`PISubscriber`

    '''
    connector_key = dict(cls='pi', type='subscribe')

    property_defs = dict(
        afelement=prop('afelement', dtype='string', required=True),
        iselementtemplate=prop('iselementtemplate', dtype='boolean', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        rmretdel=prop('rmretdel', dtype='boolear'),
        pisystem=prop('pisystem', dtype='string'),
        afdatabase=prop('afdatabase', dtype='string'),
        afrootelement=prop('afrootelement', dtype='string'),
        afattribute=prop('afattribute', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string')
    )

    def __init__(self, afelement=None, iselementtemplate=None, name=None, is_active=None,
                 snapshot=None, rmretdel=None, pisystem=None,
                 afdatabase=None, afrootelement=None, afattribute=None,
                 configfilesection=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'pi', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['afelement',
                                                   'iselementtemplate'],
                                         delete='type')
        return cls(req[0], req[1], name=name, is_active=is_active, **properties)


class PIPublisher(Connector):
    '''
    Publish operations to a PI Asset Framework (AF) server

    Parameters
    ----------
    afelement : string
        Specifies the AF element or element template name.
        Wildcards are supported.
    iselementtemplate : boolean
        Specifies that the afelement parameter is an element template name.
        By default, the afelement parameter specifies an element name.
    blocksize : int, optional
        Specifies the number of events to include in a published event
        block. The default value is 1.
    transactional : string, optional
        Sets the event block type to transactional. The default value is normal
    pisystem : string, optional
        Specifies the PI system. The default is the PI system that is
        configured in the PI AF client.
    afdatabase : string, optional
        Specifies the AF database. The default is the AF database that is
        configured in the PI AF client.
    afrootelement : string, optional
        Specifies the root element in the AF hierarchy from which to search
        for the parameter afelement. The default is the top-level element
        in the AF database.
    afattribute : string, optional
        Specifies a specific attribute in the element. The default is all
        attributes in the element.
    archivetimestamp : boolean, optional
        Specifies that all archived values from the specified timestamp
        onwards are to be published when connecting to the PI system.
        The default is to publish only new values.
    configfilesection : string, optional
        Specifies the name of the section in the config file to parse for
        configuration parameters. Specify the value as [configfilesection].
    publishwithupsert : boolean, optional
        Builds events with opcode=Upsert instead of Insert.
    maxevents : int, optional
         Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`PIPublisher`

    '''
    connector_key = dict(cls='pi', type='publish')

    property_defs = dict(
        afelement=prop('afelement', dtype='string', required=True),
        iselementtemplate=prop('iselementtemplate', dtype='boolean', required=True),
        blocksize=prop('blocksize', dtype='int'),
        transactional=prop('transactional', dtype='string'),
        pisystem=prop('pisystem', dtype='string'),
        afdatabase=prop('afdatabase', dtype='string'),
        afrootelement=prop('afrootelement', dtype='string'),
        afattribute=prop('afattribute', dtype='string'),
        archivetimestamp=prop('archivetimestamp', dtype='boolean'),
        configfilesection=prop('configfilesection', dtype='string'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, afelement=None, iselementtemplate=None, name=None, is_active=None,
                 blocksize=None, transactional=None, pisystem=None,
                 afdatabase=None, afrootelement=None, afattribute=None,
                 archivetimestamp=None, configfilesection=None,
                 publishwithupsert=None, maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'pi', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['afelement',
                                                   'iselementtemplate'],
                                         delete='type')
        return cls(req[0], req[1], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
General utilities for dealing with keywords

'''

from __future__ import print_function, division, absolute_import, unicode_literals

import keyword

DEKEYWORDS = set([(x + '_') for x in keyword.kwlist])


def dekeywordify(name):
    '''
    Add an underscore to names that are keywords

    Parameters
    ----------
    name : string
        The string to check against keywords

    Returns
    -------
    string
        Name changed to avoid keywords

    '''
    if keyword.iskeyword(name):
        return name + '_'
    return name


def keywordify(name):
    '''
    Convert name that has been dekeywordified to a keyword

    Parameters
    ----------
    name : string
        The string to convert to a keyword if needed

    Returns
    -------
    string
        Name changed to not avoid keywords

    '''
    if name in DEKEYWORDS:
        return name[:-1]
    return name



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Pylon Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class PylonPublisher(Connector):
    '''
    Publish Basler GigE camera captured frames

    Parameters
    ----------
    cameraipaddress : string, optional
         Specifies the camera IP address. The default value is the
         address of the first camera found on the local subnet.
    maxnumframes : int, optional
        Specifies the maximum number of frames to publish. The default
        value is no maximum.
    maxframerate : int, optional
        Specifies the maximum number of frames per second to publish.
        The default value is the rate at which frames are received
        from the camera.
    camerafeaturesfile : string, optional
        Specifies a Pylon Features Stream (.pfs) configuration file
        to load. The default is to use the current camera
        configuration unmodified.
    camerawidth : int, optional
        Specifies the Area-Of-Interest width. The default value is
        the value in the current camera configuration.
    cameraheight : int, optional
        Specifies the Area-Of-Interest height. The default value is
        the value in the current camera configuration.
    camerapixelformat : string, optional
        Specifies the image pixel format. The default value is the
        format in the current camera configuration.
    camerapacketsize : int, optional
        Specifies the Ethernet packet size. The default value is
        the value in the current camera configuration.
    transactional : string, optional
        Sets the event block type to transactional. The default
        value is normal.
    configfilesection : string, optional
        Specifies the name of the section in the connector config
        file to parse for configuration parameters. Specify the
        value as [configfilesection].
    publishwithupsert : boolean, optional
        Specifies to build events with opcode=Upsert instead
        of opcode=Insert
    cameraxoffset : int, optional
        Specifies the Area-Of-Interest horizontal offset. The
        default value is the value in the current camera configuration.
    camerayoffset : int, optional
        Specifies the Area-Of-Interest vertical offset. The
        default value is the value in the current camera configuration.
    maxevents : int, optional
        Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`PylonPublisher`

    '''
    connector_key = dict(cls='pylon', type='publish')

    property_defs = dict(
        cameraipaddress=prop('cameraipaddress', dtype='string'),
        maxnumframes=prop('maxnumframes', dtype='int'),
        maxframerate=prop('maxframerate', dtype='int'),
        camerafeaturesfile=prop('camerafeaturesfile', dtype='string'),
        camerawidth=prop('camerawidth', dtype='int'),
        cameraheight=prop('cameraheight', dtype='int'),
        camerapixelformat=prop('camerapixelformat', dtype='string'),
        camerapacketsize=prop('camerapacketsize', dtype='int'),
        transactional=prop('transactional', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        cameraxoffset=prop('cameraxoffset', dtype='int'),
        camerayoffset=prop('camerayoffset', dtype='int'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, name=None, is_active=None,
                 cameraipaddress=None,
                 maxnumframes=None, maxframerate=None,
                 camerafeaturesfile=None, camerawidth=None,
                 cameraheight=None, camerapixelformat=None,
                 camerapacketsize=None, transactional=None,
                 configfilesection=None, publishwithupsert=None,
                 cameraxoffset=None, camerayoffset=None,
                 maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'pylon', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        properties = map_properties(cls, properties, delete='type')
        return cls(name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import copy
import six
import unittest
from . import utils as tm
from ..config import (get_option, set_option, reset_option, describe_option, options, 
                          get_suboptions, ESPOptionError, get_default,
                          check_int, check_float, check_string, check_url, check_boolean)
from ..utils.config import subscribe, _subscribers, unsubscribe


class TestConfig(tm.TestCase):

    def setUp(self):
        reset_option()

    def tearDown(self):
        reset_option()

    def test_basic(self):
        self.assertEqual(get_option('display.show_schema'), False)
#       self.assertEqual(get_option('display.notebook.repr_html'), True)
#       self.assertEqual(get_option('display.notebook.repr_javascript'), False)

        set_option('display.show_schema', True)

        self.assertEqual(get_option('display.show_schema'), True)

        with self.assertRaises(ESPOptionError):
            options.display.show_schema = 'foo'

        options.display.show_schema = False
        self.assertEqual(options.display.show_schema, False)

        with self.assertRaises(ESPOptionError):
            options.display.show_schema = 10

        self.assertEqual(options.display.show_schema, False)

        self.assertEqual(type(options.display), type(options))

        # This key exists, but it's a level in the hierarchy, not an option
#       with self.assertRaises(ESPOptionError):
#           get_option('display.notebook.css')

        options.display.show_schema = True

        reset_option('display.show_schema')

        self.assertEqual(options.display.show_schema, False)

        with self.assertRaises(ESPOptionError):
            reset_option('display.foo')
 
        with self.assertRaises(ESPOptionError):
            reset_option('display')

    def test_shortcut_options(self):
        debug_events = get_option('debug.events')
        image_scale = get_option('display.image_scale')

        self.assertEqual(get_option('debug.events'), debug_events)
        self.assertEqual(options.debug.events, debug_events)

        options.debug.events = True

        self.assertEqual(get_option('debug.events'), True)
        self.assertEqual(options.debug.events, True)
        self.assertEqual(options.events, True)

        self.assertEqual(get_option('image_scale'), image_scale)
        self.assertEqual(get_option('display.image_scale'), image_scale)
        self.assertEqual(options.image_scale, image_scale)

        options.image_scale = 0.8
        
        self.assertEqual(get_option('image_scale'), 0.8)
        self.assertEqual(get_option('display.image_scale'), 0.8)
        self.assertEqual(options.image_scale, 0.8)

        reset_option('image_scale')

        self.assertEqual(get_option('image_scale'), 1.0)
        self.assertEqual(get_option('display.image_scale'), 1.0)
        self.assertEqual(options.image_scale, 1.0)

    def test_missing_options(self):
        with self.assertRaises(ESPOptionError):
            set_option('debug.foo', 10) 

        with self.assertRaises(ESPOptionError):
            options.debug.foo = 10

        with self.assertRaises(ESPOptionError):
            get_option('debug.foo') 

        with self.assertRaises(ESPOptionError):
            print(options.debug.foo)

        # You can not access a midpoint in the hierarchy with (s|g)et_option
        with self.assertRaises(ESPOptionError):
            set_option('debug', 10)

        with self.assertRaises(ESPOptionError):
            get_option('debug')

    def test_function_subscribers(self):
        opts = {}

        def options_subscriber(key, value, opts=opts):
            opts[key] = value

        num_subscribers = len(_subscribers) 

        subscribe(options_subscriber)

        self.assertEqual(len(_subscribers), num_subscribers + 1) 

        options.display.show_schema = True
        self.assertEqual(opts, {'display.show_schema': True})

        options.display.show_schema = False
        self.assertEqual(opts, {'display.show_schema': False})

        options.display.image_scale = 0.8
        self.assertEqual(opts, {'display.show_schema': False, 'display.image_scale': 0.8})

        options.display.image_scale = 0.8
        self.assertEqual(opts, {'display.show_schema': False, 'display.image_scale': 0.8})

        options.display.show_schema = True
        self.assertEqual(opts, {'display.show_schema': True, 'display.image_scale': 0.8})

        options.display.show_schema = False
        reset_option('display.show_schema')
        self.assertEqual(opts, {'display.show_schema': False, 'display.image_scale': 0.8})

        unsubscribe(options_subscriber)

        self.assertEqual(len(_subscribers), num_subscribers) 

        subscribe(options_subscriber)

        self.assertEqual(len(_subscribers), num_subscribers + 1) 

        del options_subscriber

        self.assertEqual(len(_subscribers), num_subscribers) 

        options.display.show_schema = False

        self.assertEqual(opts, {'display.show_schema': False, 'display.image_scale': 0.8})

    def _test_method_subscribers(self):
        opts = {}

        class OptionsSubscriber(object):
            def options_subscriber(self, key, value, opts=opts):
                opts[key] = value
        os = OptionsSubscriber()

        num_subscribers = len(_subscribers)

        subscribe(os.options_subscriber)

        self.assertEqual(len(_subscribers), num_subscribers + 1)

        options.display.show_schema = False
        self.assertEqual(opts, {'display.show_schema': False})

        options.display.show_schema = True
        self.assertEqual(opts, {'display.show_schema': True})

        options.debug.events = True
        self.assertEqual(opts, {'display.show_schema': True, 'debug.events': True})

        options.debug.events = False
        self.assertEqual(opts, {'display.show_schema': True, 'debug.events': False})

        options.display.show_schema = False
        self.assertEqual(opts, {'display.show_schema': False, 'debug.events': False})

        reset_option('display.show_schema')
        self.assertEqual(opts, {'display.show_schema': True, 'debug.events': False})

        unsubscribe(os.options_subscriber)

        self.assertEqual(len(_subscribers), num_subscribers)

        subscribe(os.options_subscriber)

        self.assertEqual(len(_subscribers), num_subscribers + 1)

        del os.options_subscriber

        self.assertEqual(len(_subscribers), num_subscribers)

        options.display.show_schema = False

        self.assertEqual(opts, {'display.show_schema': True, 'debug.events': False})

    def test_errors(self):
        with self.assertRaises(ESPOptionError):
            set_option('display.show_schema', 'hi')

    def test_doc(self):
        out = describe_option('display.show_schema', 'debug.events', _print_desc=False)
        for line in out.split('\n'):
            if not line or line.startswith(' '):
                continue
            self.assertRegex(line, r'^(display\.show_schema|debug\.events)')

        # Displays entire option hierarchy
        out = describe_option('debug', _print_desc=False)
        for line in out.split('\n'):
            if not line or line.startswith(' '):
                continue
            self.assertRegex(line, r'^debug\.')

        with self.assertRaises(ESPOptionError):
           describe_option('debug.foo')

        out = describe_option(_print_desc=False)
        self.assertRegex(out, r'\bdebug\.events :')
        self.assertRegex(out, r'\bdebug\.requests :')
        self.assertRegex(out, r'\bhostname :')
        self.assertRegex(out, r'\bport :')

    def test_suboptions(self):
        self.assertEqual(list(sorted(get_suboptions('debug').keys())), 
                          ['events', 'request_bodies', 'requests', 'responses'])

        with self.assertRaises(ESPOptionError):
            get_suboptions('display.foo')

        # This is an option, not a level in the hierarchy
        with self.assertRaises(ESPOptionError):
            get_suboptions('display.show_schema')

    def test_get_default(self):
        self.assertEqual(get_default('display.show_schema'), False)

        with self.assertRaises(ESPOptionError):
            get_default('display.foo')

        # This is a level in the hierarchy, not an option
        with self.assertRaises(ESPOptionError):
            get_default('debug')

    def test_check_int(self):
        self.assertEqual(check_int(10), 10) 
        self.assertEqual(check_int(999999999999), 999999999999) 
        self.assertEqual(check_int('10'), 10) 

        with self.assertRaises(ESPOptionError):
            check_int('foo')

        self.assertEqual(check_int(10, minimum=9), 10) 
        self.assertEqual(check_int(10, minimum=10), 10) 
        with self.assertRaises(ESPOptionError):
            check_int(10, minimum=11)
       
        self.assertEqual(check_int(10, minimum=9, exclusive_minimum=True), 10) 
        with self.assertRaises(ESPOptionError):
            check_int(10, minimum=10, exclusive_minimum=True)
        with self.assertRaises(ESPOptionError):
            check_int(10, minimum=11, exclusive_minimum=True)
       
        self.assertEqual(check_int(10, maximum=11), 10) 
        self.assertEqual(check_int(10, maximum=10), 10) 
        with self.assertRaises(ESPOptionError):
            check_int(10, maximum=9)
       
        self.assertEqual(check_int(10, maximum=11, exclusive_minimum=True), 10) 
        with self.assertRaises(ESPOptionError):
            check_int(10, maximum=10, exclusive_maximum=True)
        with self.assertRaises(ESPOptionError):
            check_int(10, maximum=9, exclusive_maximum=True)

        self.assertEqual(check_int(10, multiple_of=5), 10)
        with self.assertRaises(ESPOptionError):
            check_int(10, multiple_of=3) 

    def test_check_float(self):
        self.assertEqual(check_float(123.567), 123.567)
        self.assertEqual(check_float(999999999999.999), 999999999999.999)
        self.assertEqual(check_float('123.567'), 123.567)

        with self.assertRaises(ESPOptionError):
            check_float('foo')

        self.assertEqual(check_float(123.567, minimum=123.566), 123.567)
        self.assertEqual(check_float(123.567, minimum=123.567), 123.567)
        with self.assertRaises(ESPOptionError):
            check_float(123.567, minimum=123.577)

        self.assertEqual(check_float(123.567, minimum=123.566, exclusive_minimum=True), 123.567)
        with self.assertRaises(ESPOptionError):
            check_float(123.567, minimum=123.567, exclusive_minimum=True)
        with self.assertRaises(ESPOptionError):
            check_float(123.567, minimum=123.568, exclusive_minimum=True)

        self.assertEqual(check_float(123.567, maximum=123.568), 123.567)
        self.assertEqual(check_float(123.567, maximum=123.567), 123.567)
        with self.assertRaises(ESPOptionError):
            check_float(123.567, maximum=123.566)

        self.assertEqual(check_float(123.567, maximum=123.567, exclusive_minimum=True), 123.567)
        with self.assertRaises(ESPOptionError):
            check_float(123.567, maximum=123.567, exclusive_maximum=True)
        with self.assertRaises(ESPOptionError):
            check_float(123.567, maximum=123.566, exclusive_maximum=True)

        with self.assertRaises(ESPOptionError):
            check_float(123.567, multiple_of=3)

    def test_check_string(self):
        self.assertEqual(check_string('hi there'), 'hi there')
        self.assertTrue(isinstance(check_string('hi there'), six.string_types))

        self.assertEqual(check_string('hi there', pattern=r' th'), 'hi there')
        with self.assertRaises(ESPOptionError):
            check_string('hi there', pattern=r' th$')

        self.assertEqual(check_string('hi there', max_length=20), 'hi there')
        self.assertEqual(check_string('hi there', max_length=8), 'hi there')
        with self.assertRaises(ESPOptionError):
            check_string('hi there', max_length=7)

        self.assertEqual(check_string('hi there', min_length=3), 'hi there')
        self.assertEqual(check_string('hi there', min_length=8), 'hi there')
        with self.assertRaises(ESPOptionError):
            check_string('hi there', min_length=9)

        self.assertEqual(check_string('hi there', valid_values=['hi there', 'bye now']), 'hi there')
        with self.assertRaises(ESPOptionError):
            check_string('foo', valid_values=['hi there', 'bye now']) 

        # Invalid utf8 data
        with self.assertRaises(ESPOptionError):
            check_string(b'\xff\xfeW[')

    def test_check_url(self):
        self.assertEqual(check_url('hi there'), 'hi there')
        self.assertTrue(isinstance(check_url('hi there'), six.string_types))

        # Invalid utf8 data
        with self.assertRaises(ESPOptionError):
            check_url(b'\xff\xfeW[')

    def test_check_boolean(self):
        self.assertEqual(check_boolean(True), True)
        self.assertEqual(check_boolean(False), False)
        self.assertEqual(check_boolean(1), True)
        self.assertEqual(check_boolean(0), False)

        with self.assertRaises(ESPOptionError):
            check_boolean(2)
        with self.assertRaises(ESPOptionError):
            check_boolean('true')
        with self.assertRaises(ESPOptionError):
            check_boolean(1.1)


if __name__ == '__main__':
    tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
Options interface for ESP

Options can be set and retrieved using set_option(...), get_option(...), and
reset_option(...).  The describe_option(...) function can be used to display
a description of one or more options.

'''

from __future__ import print_function, division, absolute_import, unicode_literals

import contextlib
import os
import re
import six
import types
import weakref
from six.moves.urllib.parse import urlparse
from .xdict import xdict
from ..exceptions import ESPOptionError

# pylint: disable=C0330


# Container for options
_config = xdict()

# Subscribers to option changes
_subscribers = weakref.WeakKeyDictionary()


def iteroptions(*args, **kwargs):
    '''
    Iterate through name / value pairs of options

    Options can come in several forms.  They can be consecutive arguments
    where the first argument is the name and the following argument is
    the value.  They can be two-element tuples (or lists) where the first
    element is the name and the second element is the value.  You can
    also pass in a dictionary of key / value pairs.  And finally, you can
    use keyword arguments.

    Parameters
    ----------
    *args : any, optional
        See description above.
    **kwargs : key / value pairs, optional
        Arbitrary keyword arguments.

    Returns
    -------
    generator
        Each iteration returns a name / value pair in a tuple

    '''
    args = list(args)
    while args:
        item = args.pop(0)
        if isinstance(item, (list, tuple)):
            yield item[0], item[1]
        elif isinstance(item, dict):
            for key, value in six.iteritems(item):
                yield key, value
        else:
            yield item, args.pop(0)
    for key, value in six.iteritems(kwargs):
        yield key, value


def subscribe(func):
    '''
    Add a subscriber function to option events

    Parameters
    ----------
    func : callable
        A callable object that takes two parameters: key and value.
        This function is called with the name and value of any option
        that is set.

    Returns
    -------
    None

    '''
    if isinstance(func, types.MethodType):
        obj = six.get_method_self(func)
        func = six.get_method_function(func)
        _subscribers[func] = (weakref.ref(func), weakref.ref(obj))
    else:
        _subscribers[func] = (weakref.ref(func), None)


def unsubscribe(func):
    '''
    Remove a subscriber from option events

    Parameters
    ----------
    func : callable
        The callable used to subscribe to option events

    Returns
    -------
    None

    '''
    _subscribers.pop(func, None)


@contextlib.contextmanager
def option_context(*args, **kwargs):
    '''
    Create a context for setting option temporarily

    Parameters
    ----------
    *args : string / any pairs
        Name / value pairs in consecutive arguments (not tuples)
    **kwargs : dict
        Key / value pairs of options

    '''
    # Save old state and set new option values
    oldstate = {}
    for key, value in iteroptions(*args, **kwargs):
        key = key.lower()
        oldstate[key] = get_option(key)
        set_option(key, value)

    # Yield control
    yield

    # Set old state back
    for key, value in six.iteritems(oldstate):
        set_option(key, value)


def _get_option_leaf_node(key):
    '''
    Find full option name of given key

    Parameters
    ----------
    key : string
        Either a partial key or full key name of an option

    Returns
    -------
    string
        The full key name of the option

    Raises
    ------
    ESPOptionError
        If more than one option matches

    '''
    flatkeys = list(_config.flatkeys())
    key = key.lower()
    if key in flatkeys:
        return key
    keys = [k for k in flatkeys if k.endswith('.' + key)]
    if len(keys) > 1:
        raise ESPOptionError('There is more than one option with the name %s.' % key)
    if not keys:
        raise ESPOptionError('%s is not a valid option name.' % key)
    return keys[0]


def set_option(*args, **kwargs):
    '''
    Set the value of an option

    Parameters
    ----------
    *args : string / any pairs
        The name and value of an option in consecutive arguments (not tuples)
    **kwargs : dict
        Arbitrary keyword / value pairs

    '''
    for key, value in iteroptions(*args, **kwargs):
        key = _get_option_leaf_node(key)
        opt = _config[key]
        if not isinstance(opt, SWATOption):
            raise ESPOptionError('%s is not a valid option name' % key)
        opt.set(value)


set_options = set_option


def get_option(key):
    '''
    Get the value of an option

    Parameters
    ----------
    key : string
        The name of the option

    Returns
    -------
    any
        The value of the option

    '''
    key = _get_option_leaf_node(key)
    opt = _config[key]
    if not isinstance(opt, SWATOption):
        raise ESPOptionError('%s is not a valid option name' % key)
    return opt.get()


def get_suboptions(key):
    '''
    Get the dictionary of options at the level `key`

    Parameters
    ----------
    key : string
        The name of the option collection

    Returns
    -------
    dict
        The dictionary of options at level `key`

    '''
    if key not in _config:
        raise ESPOptionError('%s is not a valid option name' % key)
    opt = _config[key]
    if isinstance(opt, SWATOption):
        raise ESPOptionError('%s does not have sub-options' % key)
    return opt


def get_default(key):
    '''
    Get the default value of an option

    Parameters
    ----------
    key : string
        The name of the option

    Returns
    -------
    any
        The default value of the option

    '''
    key = _get_option_leaf_node(key)
    opt = _config[key]
    if not isinstance(opt, SWATOption):
        raise ESPOptionError('%s is not a valid option name' % key)
    return opt.get_default()


get_default_val = get_default


def describe_option(*keys, **kwargs):
    '''
    Print the description of one or more options

    Parameters
    ----------
    *keys : one or more strings
        Names of the options

    '''
    _print_desc = kwargs.get('_print_desc', True)

    out = []

    if not keys:
        keys = sorted(_config.flatkeys())
    else:
        newkeys = []
        for k in keys:
            try:
                newkeys.append(_get_option_leaf_node(k))
            except ESPOptionError:
                newkeys.append(k)

    for key in keys:

        if key not in _config:
            raise ESPOptionError('%s is not a valid option name' % key)

        opt = _config[key]
        if isinstance(opt, xdict):
            desc = describe_option(*['%s.%s' % (key, x)
                                   for x in opt.flatkeys()], _print_desc=_print_desc)
            if desc is not None:
                out.append(desc)
            continue

        if _print_desc:
            print(opt.__doc__)
            print('')
        else:
            out.append(opt.__doc__)

    if not _print_desc:
        return '\n'.join(out)


def reset_option(*keys):
    '''
    Reset one or more options back to their default value

    Parameters
    ----------
    *keys : one or more strings
        Names of options to reset

    '''
    if not keys:
        keys = sorted(_config.flatkeys())
    else:
        keys = [_get_option_leaf_node(k) for k in keys]

    for key in keys:

        if key not in _config:
            raise ESPOptionError('%s is not a valid option name' % key)

        opt = _config[key]
        if not isinstance(opt, SWATOption):
            raise ESPOptionError('%s is not a valid option name' % key)

        # Reset swat options
        set_option(key, get_default(key))


def check_int(value, minimum=None, maximum=None, exclusive_minimum=False,
              exclusive_maximum=False, multiple_of=None):
    '''
    Validate an integer value

    Parameters
    ----------
    value : int or float
        Value to validate
    minimum : int, optional
        The minimum value allowed
    maximum : int, optional
        The maximum value allowed
    exclusive_minimum : boolean, optional
        Should the minimum value be excluded as an endpoint?
    exclusive_maximum : boolean, optional
        Should the maximum value be excluded as an endpoint?
    multiple_of : int, optional
        If specified, the value must be a multple of it in order for
        the value to be considered valid.

    Returns
    -------
    int
        The validated integer value

    '''
    try:
        out = int(value)
    except:
        raise ESPOptionError('Could not convert %s to an integer' % value)

    if minimum is not None:
        if out < minimum:
            raise ESPOptionError('%s is smaller than the minimum value of %s' %
                                 (out, minimum))
        if exclusive_minimum and out == minimum:
            raise ESPOptionError('%s is equal to the exclusive nimum value of %s' %
                                 (out, minimum))

    if maximum is not None:
        if out > maximum:
            raise ESPOptionError('%s is larger than the maximum value of %s' %
                                 (out, maximum))
        if exclusive_maximum and out == maximum:
            raise ESPOptionError('%s is equal to the exclusive maximum value of %s' %
                                 (out, maximum))

    if multiple_of is not None and (out % int(multiple_of)) != 0:
        raise ESPOptionError('%s is not a multiple of %s' % (out, multiple_of))

    return out


def check_float(value, minimum=None, maximum=None, exclusive_minimum=False,
                exclusive_maximum=False, multiple_of=None):
    '''
    Validate a floating point value

    Parameters
    ----------
    value : int or float
        Value to validate
    minimum : int or float, optional
        The minimum value allowed
    maximum : int or float, optional
        The maximum value allowed
    exclusive_minimum : boolean, optional
        Should the minimum value be excluded as an endpoint?
    exclusive_maximum : boolean, optional
        Should the maximum value be excluded as an endpoint?
    multiple_of : int or float, optional
        If specified, the value must be a multple of it in order for
        the value to be considered valid.

    Returns
    -------
    float
        The validated floating point value

    '''
    try:
        out = float(value)
    except:
        raise ESPOptionError('Could not convert %s to a float' % value)

    if minimum is not None:
        if out < minimum:
            raise ESPOptionError('%s is smaller than the minimum value of %s' %
                                 (out, minimum))
        if exclusive_minimum and out == minimum:
            raise ESPOptionError('%s is equal to the exclusive nimum value of %s' %
                                 (out, minimum))

    if maximum is not None:
        if out > maximum:
            raise ESPOptionError('%s is larger than the maximum value of %s' %
                                 (out, maximum))
        if exclusive_maximum and out == maximum:
            raise ESPOptionError('%s is equal to the exclusive maximum value of %s' %
                                 (out, maximum))

    if multiple_of is not None and (out % int(multiple_of)) != 0:
        raise ESPOptionError('%s is not a multiple of %s' % (out, multiple_of))

    return out


def check_boolean(value):
    '''
    Validate a boolean value

    Parameters
    ----------
    value : int or boolean
        The value to validate.  If specified as an integer, it must
        be either 0 for False or 1 for True.

    Returns
    -------
    boolean
        The validated boolean

    '''
    if value is False or value is True:
        return value

    if isinstance(value, int):
        if value == 1:
            return True
        if value == 0:
            return False

    raise ESPOptionError('%s is not a boolean or proper integer value')


def check_string(value, pattern=None, max_length=None, min_length=None,
                 valid_values=None):
    '''
    Validate a string value

    Parameters
    ----------
    value : string
        The value to validate
    pattern : regular expression string, optional
        A regular expression used to validate string values
    max_length : int, optional
        The maximum length of the string
    min_length : int, optional
        The minimum length of the string
    valid_values : list of strings, optional
        List of the only possible values

    Returns
    -------
    string
        The validated string value

    '''
    try:
        if isinstance(value, six.text_type):
            out = value
        elif isinstance(value, six.binary_type):
            out = value.decode('utf-8')
        else:
            out = '%s' % value
    except:
        raise ESPOptionError('Could not convert string value to unicode')

    if max_length is not None and len(out) > max_length:
        raise ESPOptionError('%s is longer than the maximum length of %s' %
                             (out, max_length))

    if min_length is not None and len(out) < min_length:
        raise ESPOptionError('%s is shorter than the minimum length of %s' %
                             (out, min_length))

    if pattern is not None and not re.search(pattern, out):
        raise ESPOptionError('%s does not match pattern %s' % (out, pattern))

    if valid_values is not None and out not in valid_values:
        raise ESPOptionError('%s is not one of the possible values: %s' %
                             (out, ', '.join(valid_values)))

    return out


def check_url(value, pattern=None, max_length=None, min_length=None, valid_values=None):
    '''
    Validate a URL value

    Parameters
    ----------
    value : any
        The value to validate.  This value will be cast to a string
        and converted to unicode.
    pattern : regular expression string, optional
        A regular expression used to validate string values
    max_length : int, optional
        The maximum length of the string
    min_length : int, optional
        The minimum length of the string
    valid_values : list of strings, optional
        List of the only possible values

    Returns
    -------
    string
        The validated URL value

    '''
    out = check_string(value, pattern=pattern, max_length=max_length,
                       min_length=min_length, valid_values=valid_values)
    try:
        urlparse(out)
    except:
        raise ESPOptionError('%s is not a valid URL' % value)
    return out


class SWATOption(object):
    '''
    SWAT configuration option

    Parameters
    ----------
    name : string
        The name of the option
    typedesc : string
        Description of the option data type (e.g., int, float, string)
    validator : callable
        A callable object that validates the option value and returns
        the validated value.
    default : any
        The default value of the option
    doc : string
        The documentation string for the option

    environ : string, optional
        If specified, the value should be specified in an environment
        variable of that name.

    Returns
    -------
    SWATOption object

    '''

    def __init__(self, name, typedesc, validator, default, doc, environ=None):
        self._name = name
        self._typedesc = typedesc
        self._validator = validator
        if environ is not None:
            self._default = validator(os.environ.get(environ, default))
        else:
            self._default = validator(default)
        self._environ = environ
        self._value = self._default
        self._doc = doc

    @property
    def __doc__(self):
        ''' Documentation string '''
        separator = ' '
        if isinstance(self._value, six.string_types) and len(self._value) > 40:
            separator = '\n    '
        return '''%s : %s\n    %s\n    [default: %s]%s[currently: %s]\n''' % \
            (self._name, self._typedesc, self._doc.rstrip().replace('\n', '\n    '),
             self._default, separator, self._value)

    def set(self, value):
        '''
        Set the value of the option

        Parameters
        ----------
        value : any
           The value to set

        Returns
        -------
        None

        '''
        value = self._validator(value)
        _config[self._name]._value = value

        if self._environ is not None:
            os.environ[self._environ] = str(value)

        for func, obj in list(_subscribers.values()):
            if func is not None:
                if obj is None:
                    func = func()
                    if func is not None:
                        func(self._name, value)
                else:
                    func, obj = func(), obj()
                    if func is not None and obj is not None:
                        func(obj, self._name, value)

    def get(self):
        '''
        Get the value of the option

        Returns
        -------
        any
            The value of the option

        '''
        if self._environ is not None:
            try:
                _config[self._name]._value = self._validator(os.environ[self._environ])
            except KeyError:
                pass
        return _config[self._name]._value

    def get_default(self):
        '''
        Get the default value of the option

        Returns
        -------
        any
            The default value of the option

        '''
        return _config[self._name]._default


def register_option(key, typedesc, validator, default, doc, environ=None):
    '''
    Register a new option

    Parameters
    ----------
    key : string
        The name of the option
    typedesc : string
        Description of option data type (e.g., int, float, string)
    validator : callable
        A callable object that validates the value and returns
        a validated value.
    default : any
        The default value of the option
    doc : string
        The documentation string for the option

    environ : string, optional
        If specified, the value should be specified in an environment
        variable of that name.

    Returns
    -------
    None

    '''
    _config[key] = SWATOption(key, typedesc, validator, default, doc, environ=environ)


class AttrOption(object):
    '''
    Attribute-style access of SWAT options

    '''

    def __init__(self, name):
        object.__setattr__(self, '_name', name)

    def __dir__(self):
        if self._name in _config:
            return _config[self._name].flatkeys()
        return _config.flatkeys()

    @property
    def __doc__(self):
        if self._name:
            return describe_option(self._name, _print_desc=False)
        return describe_option(_print_desc=False)

    def __getattr__(self, name):
        name = name.lower()
        if self._name:
            fullname = self._name + '.' + name
        else:
            fullname = name
        if fullname not in _config:
            fullname = _get_option_leaf_node(fullname)
        out = _config[fullname]
        if not isinstance(out, SWATOption):
            return type(self)(fullname)
        return out.get()

    def __setattr__(self, name, value):
        name = name.lower()
        if self._name:
            fullname = self._name + '.' + name
        else:
            fullname = name
        if fullname not in _config:
            fullname = _get_option_leaf_node(fullname)
        out = _config[fullname]
        if not isinstance(out, SWATOption):
            return type(self)(fullname)
        _config[fullname].set(value)
        return

    def __call__(self, *args, **kwargs):
        ''' Shortcut for option context '''
        return option_context(*args, **kwargs)


# Object for setting and getting options using attribute syntax
options = AttrOption(None)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import datetime
import esppy
import os
import six
import sys
import unittest
from esppy.utils.xml import (new_elem, add_elem, add_properties, from_xml,
                             to_xml, xml_indent, get_attrs)
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestXMLUtils(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def test_new_elem(self):
        # empty
        elem = new_elem('foo')
        self.assertEqual(to_xml(elem), '<foo />')

        # With attributes
        elem = new_elem('foo', attrib=dict(a='100', b='200'))
        self.assertEqual(to_xml(elem), '<foo a="100" b="200" />')

        # With text content
        elem = new_elem('foo', attrib=dict(a='100', b='200'), text_content='Hi')
        self.assertEqual(to_xml(elem), '<foo a="100" b="200">Hi</foo>')

        # With keyword parameters
        elem = new_elem('foo', attrib=dict(a='100', b='200'), other='value')
        self.assertEqual(to_xml(elem), '<foo a="100" b="200" other="value" />')

    def test_add_elem(self):
        # tag name
        parent = new_elem('foo')
        add_elem(parent, 'bar', a='100', b='200', attrib=dict(c="300"))
        self.assertEqual(to_xml(parent), '<foo><bar a="100" b="200" c="300" /></foo>')

        # element
        parent = new_elem('foo')
        add_elem(parent, new_elem('bar', a='100', b='200'), attrib=dict(c="300"), other='value')
        self.assertEqual(to_xml(parent), '<foo><bar a="100" b="200" c="300" other="value" /></foo>')

        # XML
        parent = new_elem('foo')
        add_elem(parent, '<bar b="200" a="100" />', attrib=dict(c="300"), other='value')
        self.assertEqual(to_xml(parent), '<foo><bar a="100" b="200" c="300" other="value" /></foo>')

    def test_xml_indent(self):
        a = new_elem('a', x='100', y='200')
        b = new_elem('b', m='1000', n='2000')
        c = new_elem('c', text_content='Hi')
        add_elem(a, b)
        add_elem(b, c)
        xml_indent(a)
        self.assertEqual(to_xml(a), '<a x="100" y="200">\n  <b m="1000" n="2000">\n    <c>Hi</c>\n  </b>\n</a>\n')

    def test_to_xml(self):
        a = new_elem('a', x='100', y='200')
        b = new_elem('b', m='1000', n='2000')
        c = new_elem('c', text_content='Hi')
        add_elem(a, b)
        add_elem(b, c)
        self.assertEqual(to_xml(a), '<a x="100" y="200"><b m="1000" n="2000"><c>Hi</c></b></a>')

        a = new_elem('a', x='100', y='200')
        b = new_elem('b', m='1000', n='2000')
        c = new_elem('c', text_content='Hi')
        add_elem(a, b)
        add_elem(b, c)
        self.assertEqual(to_xml(a, pretty=True), '<a x="100" y="200">\n  <b m="1000" n="2000">\n    <c>Hi</c>\n  </b>\n</a>\n')

        a = new_elem('a', x='100', y='200')
        b = new_elem('b', m='1000', n='2000')
        c = new_elem('c', text_content='Hi')
        add_elem(a, b)
        add_elem(b, c)
        self.assertEqual(to_xml(to_xml(a), pretty=True), '<a x="100" y="200">\n  <b m="1000" n="2000">\n    <c>Hi</c>\n  </b>\n</a>\n')

        a = new_elem('a', x='100', y='200')
        b = new_elem('b', m='1000', n='2000')
        c = new_elem('c', text_content='Hi')
        add_elem(a, b)
        add_elem(b, c)
        print(to_xml(a, encoding='ascii'))
        self.assertEqual(to_xml(a, encoding='ascii'), b'<?xml version=\'1.0\' encoding=\'ascii\'?>\n<a x="100" y="200"><b m="1000" n="2000"><c>Hi</c></b></a>')

    @unittest.skip('Test is obsolete; needs to be rewritten for "xml_map" and "attribute"s')
    def test_get_attrs(self):
        project = self.project

        class XMLAttrObject1(object):
            attr_map = ['a', 'b', 'c']
            def __init__(self):
                self.a = 1
                self.b = 2
                self.c = 3
                self.d = 4
                self.e = 5

        class XMLAttrObject2(object):
            attr_map = {'a': 'attr1', 'b': 'attr2', 'c': None}
            def __init__(self):
                self.a = 100
                self.b = 200
                self.c = 300

        class XMLAttrObject3(object):
            attr_map = 10

        class XMLAttrObject4(object):
            def __init__(self):
                self.a = 1000
                self.b = 2000
                self.c = 3000
                self.d = None 
                self.e = [1, 2, 3]
                self.f = {'key': 'value'}
                self.g = True
                self.h = False
                self._i = 10
                self.cls = 20
                self.my_project = project

        x1 = XMLAttrObject1()
        x2 = XMLAttrObject2()
        x3 = XMLAttrObject3()
        x4 = XMLAttrObject4()

        out = get_attrs(x1)
        self.assertEqual(out, {'a': '1', 'b': '2', 'c': '3'})

        out = get_attrs(x1, exclude='a')
        self.assertEqual(out, {'b': '2', 'c': '3'})

        out = get_attrs(x1, extra='d')
        self.assertEqual(out, {'a': '1', 'b': '2', 'c': '3', 'd': '4'})

        out = get_attrs(x1, extra=['d', 'e'])
        self.assertEqual(out, {'a': '1', 'b': '2', 'c': '3', 'd': '4', 'e': '5'})

        out = get_attrs(x2)
        self.assertEqual(out, {'attr1': '100', 'attr2': '200', 'c': '300'})

        with self.assertRaises(TypeError):
            get_attrs(x3)

        out = get_attrs(x4)
        self.assertEqual(out, {'a': '1000', 'b': '2000', 'c': '3000', 'g': 'true',
                               'h': 'false', 'class': '20',
                               'my-project': 'ESPUnitTestProjectSA'})

        out = get_attrs(x4, exclude=['c', 'd', 'e', 'f', 'g', 'h'])
        self.assertEqual(out, {'a': '1000', 'b': '2000', 'class': '20',
                               'my-project': 'ESPUnitTestProjectSA'})

if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os
import unittest
import esppy
from . import utils as tm


class TestCalculateWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('calculate_window.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest',
                                  pubsub='auto', n_threads='1', use_tagged_token=True)

        cq = esp.ContinuousQuery(name='cq_01',trace='w_calculate')

        src = esp.SourceWindow(name='src_win',
                               schema=('id*:int64','x_c:double','y_c:double'),
                               insert_only=True, autogen_key=False)
        src.add_connector('fs', conn_name='pub', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='calculate_window.csv',
                          transactional=True,
                          blocksize='1')

        calculate_win = esp.CalculateWindow(name='calculate_win',
                                            algorithm='Correlation',
                                            schema=('id*:int64','y_c:double',
                                                    'x_c:double','corOut:double'))
        calculate_win.set_parameters(windowLength='5')
        calculate_win.set_inputs(x='x_c',y='y_c')
        calculate_win.set_outputs(corOut='corOut')
        calculate_win.add_connector('fs',
                                   conn_name='sub', conn_type='subscribe',
                                   type='sub',
                                   fstype='csv',
                                   fsname='result.out',
                                   snapshot=True)

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(calculate_win)

        src.add_target(calculate_win,role='data')

        self._test_model_file('calculate_window.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Window Transformers '''

from __future__ import print_function, division, absolute_import, unicode_literals

import io
import numpy as np
import pandas as pd
import re
import six
from PIL import Image


def _bgr2rgb(pil_image):
    return Image.fromarray(np.asarray(pil_image)[:,:,::-1])


def bgr2rgb(data, columns=None):
    '''
    Convert BGR images to RGB

    Parameters
    ----------
    data : PIL.Image or DataFrame
        The image data
    columns : string or list-of-strings, optional
        If `data` is a DataFrame, this is the list of columns that 
        contain image data.

    Returns
    -------
    :class:`PIL.Image`
        If `data` is a :class:`PIL.Image` 
    :class:`pandas.DataFrame`
        If `data` is a :class:`pandas.DataFrame`

    '''
    if hasattr(data, 'columns'):
        if len(data):
            if not columns:
                columns = list(data.columns)
            elif isinstance(columns, six.string_types):
                columns = [columns]
            for col in columns:
                if Image.isImageType(data[col].iloc[0]):
                    data[col] = data[col].apply(_bgr2rgb)
        return data

    elif Image.isImageType(data):
        return _bgr2rgb(data)

    return data


def rgb2bgr(data, columns=None):
    '''
    Convert RGB images to BGR

    Parameters
    ----------
    data : PIL.Image or DataFrame
        The image data
    columns : string or list-of-strings, optional
        If `data` is a DataFrame, this is the list of columns that
        contain image data.

    Returns
    -------
    :class:`PIL.Image`
        If `data` is a :class:`PIL.Image`
    :class:`pandas.DataFrame`
        If `data` is a :class:`pandas.DataFrame`

    '''
    return bgr2rgb(data, columns=columns)


def _bytes2image(data):
    return Image.open(io.BytesIO(data))


def bytes2image(data, columns=None):
    '''
    Convert bytes to PIL.Image objects

    Parameters
    ----------
    data : PIL.Image or DataFrame
        The image data
    columns : string or list-of-strings, optional
        If `data` is a DataFrame, this is the list of columns that
        contain image data.

    Returns
    -------
    :class:`PIL.Image`
        If `data` is a :class:`PIL.Image`
    :class:`pandas.DataFrame`
        If `data` is a :class:`pandas.DataFrame`

    '''
    if hasattr(data, 'columns'):
        if len(data):
            if not columns:
                columns = list(data.columns)
            elif isinstance(columns, six.string_types):
                columns = [columns]
            for col in columns:
                if isinstance(data[col].iloc[0], bytes): 
                    data[col] = data[col].apply(_bytes2image)
        return data

    elif isinstance(data, six.binary_type):
        return _bytes2image(data)

    return data



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

import six
from .base import Window, attribute
from .utils import ensure_element, get_args, connectors_to_end
from ..utils import xml


class Geometry(object):
    '''
    Geofence geometry

    Parameters
    ----------
    desc_fieldname : string, optional
        Specifies the name of the geometry input window’s field that
        contains the polygon coordinates data.
    data_fieldname : string, optional
        Specifies the name of the geometry input window’s field that
        contains the geometry description.
    x_fieldname : string, optional
        Specifies the name of the geometry input window’s field that
        contains the location X or longitude coordinate of the
        circle's center.
    y_fieldname : string, optional
        Specifies the name of the geometry input window’s field that
        contains the location Y or latitude coordinate of the
        circle's center.
    radius_fieldname : string, optional
        Specifies the name of the geometry input window’s field that
        contains the circle radius distance.
    radius : int or float, optional
        Specifies the default circle's radius distance.
        Default: 1000
    data_separator : string, optional
        Specifies the coordinate delimiter character used in the
        geometry data field specified by the property data-fieldname.
        Default: <space>

    Returns
    -------
    :class:`Geometry`

    '''

    def __init__(self, desc_fieldname=None, data_fieldname=None, x_fieldname=None,
                 y_fieldname=None, radius_fieldname=None, radius=None,
                 data_separator=None):
        self.desc_fieldname = desc_fieldname
        self.data_fieldname = data_fieldname
        self.x_fieldname = x_fieldname
        self.y_fieldname = y_fieldname
        self.radius_fieldname = radius_fieldname
        self.radius = radius
        self.data_separator = data_separator

    def copy(self, deep=False):
        return type(self)(desc_fieldname=self.desc_fieldname,
                          data_fieldname=self.data_fieldname,
                          x_fieldname=self.x_fieldname,
                          y_fieldname=self.y_fieldname,
                          radius_fieldname=self.radius_fieldname,
                          radius=self.radius,
                          data_separator=self.data_separator)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`Geometry`

        '''
        data = ensure_element(data)
        out = cls()
        for key, value in six.iteritems(data.attrib):
            key = key.replace('-', '_')
            if hasattr(out, key):
                setattr(out, key, value)
        return out

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('geometry',
                            attrib=dict(desc_fieldname=self.desc_fieldname,
                                        data_fieldname=self.data_fieldname,
                                        x_fieldname=self.x_fieldname,
                                        y_fieldname=self.y_fieldname,
                                        radius_fieldname=self.radius_fieldname,
                                        radius=self.radius,
                                        data_separator=self.data_separator))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class Position(object):
    '''
    Geofence Position

    Parameters
    ----------
    x_fieldname : string, optional
       Specifies the name of the position input window’s field that
       contains the position X or longitude coordinate.
    y_fieldname : string, optional
       Specifies the name of the position input window’s field that
       contains the position Y or latitude coordinate.
    lookupdistance_fieldname : string, optional
       Specifies the name of the position input window’s field that
       contains the position lookup distance.
    lookupdistance : int or float, optional
       This distance is in units for Cartesian coordinates and in
       meters for geographic coordinates.
       Default: 0.

    Returns
    -------
    :class:`Position`

    '''

    def __init__(self, x_fieldname=None, y_fieldname=None,
                 lookupdistance_fieldname=None, lookupdistance=None):
        self.x_fieldname = x_fieldname
        self.y_fieldname = y_fieldname
        self.lookupdistance_fieldname = lookupdistance_fieldname
        self.lookupdistance = lookupdistance

    def copy(self, deep=False):
        return type(self)(x_fieldname=self.x_fieldname, y_fieldname=self.y_fieldname,
                          lookupdistance_fieldname=self.lookupdistance_fieldname,
                          lookupdistance=self.lookupdistance)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`Position`

        '''
        data = ensure_element(data)
        out = cls()
        for key, value in six.iteritems(data.attrib):
            key = key.replace('-', '_')
            if hasattr(out, key):
                setattr(out, key, value)
        return out

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('position',
                            attrib=dict(x_fieldname=self.x_fieldname,
                                        y_fieldname=self.y_fieldname,
                                        loopupdistance_fieldname=self.lookupdistance_fieldname,
                                        lookupdistance=self.lookupdistance))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class Output(object):
    '''
    Geofence Output

    Parameters
    ----------
    geotype_fieldname : string, optional
        Specifies the name of the output schema field that receives
        the geometry Type.
    geoid_fieldname : string, optional
        Specifies the name of the output schema field that receives
        the geometry ID.
    geodesc_fieldname : string, optional
        Specifies the name of the output schema field that receives
        the geometry description.
    geodistance_fieldname : string, optional
        Specifies the name of the output schema field that receives
        the distance between the position and the geometry (center
        point for circle geometries and centroid for polygons).
    eventnumber_fieldname : string, optional
        Specifies the name of the output schema additional key field
        that receives the generated event number.

    Returns
    -------
    :class:`Output`

    '''

    def __init__(self, geotype_fieldname=None, geoid_fieldname=None,
                 geodesc_fieldname=None, geodistance_fieldname=None,
                 eventnumber_fieldname=None):
        self.geotype_fieldname = geotype_fieldname
        self.geoid_fieldname = geoid_fieldname
        self.geodesc_fieldname = geodesc_fieldname
        self.geodistance_fieldname = geodistance_fieldname
        self.eventnumber_fieldname = eventnumber_fieldname

    def copy(self, deep=False):
        return type(self)(geotype_fieldname=self.geotype_fieldname,
                          geoid_fieldname=self.geoid_fieldname,
                          geodesc_fieldname=self.geodesc_fieldname,
                          geodistance_fieldname=self.geodistance_fieldname,
                          eventnumber_fieldname=self.eventnumber_fieldname)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`Output`

        '''
        data = ensure_element(data)
        out = cls()
        for key, value in six.iteritems(data.attrib):
            key = key.replace('-', '_')
            if hasattr(out, key):
                setattr(out, key, value)
        return out

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('output',
                            attrib=dict(geotype_fieldname=self.geotype_fieldname,
                                        geoid_fieldname=self.geoid_fieldname,
                                        geodesc_fieldname=self.geodesc_fieldname,
                                        geodistance_fieldname=self.geodistance_fieldname,
                                        eventnumber_fieldname=self.eventnumber_fieldname))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class GeofenceWindow(Window):
    '''
    Geofence window

    Parameters
    ----------
    name : string, optional
        The name of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level
        value of pubsub is manual, true enables publishing and subscribing
        for the window and false disables it.
    description : string, optional
        Description of the window
    output_insert_only : bool, optional
        When true, prevents the window from passing non-insert events to
        other windows.
    collapse_updates : bool, optional
        When true, multiple update blocks are collapsed into a single update block
    pulse_interval : string, optional
        Output a canonical batch of updates at the specified interval
    exp_max_string : int, optional
        Specifies the maximum size of strings that the expression engine
        uses for the window. Default value is 1024.
    index_type : string, optional
        Index type for the window
    pubsub_index_type : string, optional
        Publish/subscribe index type.  Valid values are the same as for the
        `index_type` parameter.
    pubsub_index_type : string, optional
        Publish/subscribe index type.  Valid values are the same as for the
        `index_type` parameter.
    polygon_compute_distance_to : string, optional
        Specifies whether to compute the distance from the position to the
        closest segment or to the centroid.
    proximity_analysis : bool, optional
        Specifies to return polygons that are within the distance define by
        the radius property value.
        Default: false.
    geofencing_algorithm : string, optional
        Specifies the geofencing algorithm to use for polygon geometries.
        Valid Values: crossing or winding
    coordinate_type : string, optional
        Coordinate type.
        Valid Values: cartesian or geographic
        Default: cartesian
    autosize_mesh : bool, optional
        Specifies whether to compute and set the mesh factor automatically
    meshfactor_x : int, optional
        Specifies the mesh factor for the X or longitude axis.
        Default: 0
    meshfactor_y : int, optional
        Specifies the mesh factor for the Y or latitude axis.
        Default: 0
    max_meshcells_per_geometry : int, optional
        Specifies the maximum allowed mesh cells created per geometries to
        avoid creating an oversized mesh that would generate useless
        intensive calculations.
        Default: 500
    output_multiple_results : bool, optional
        Specifies whether to write a single or multiple geometries,
        regardless of the number of matching geometries.
    output_sorted_results : bool, optional
        When set to true, specifies to sort the output result by increasing
        distance between the position and the geometry.
        Default: false
    log_invalid_geometry : bool, optional
        Specifies whether to log invalid geometries in the standard
        output log.
        Default: false

    Attributes
    ----------
    geometry : Geometry
        Geofence window geometry
    position : Position
        Geofence window position
    output : Output
        Geofence window output

    Returns
    -------
    :class:`GeofenceWindow`

    '''

    window_type = 'geofence'

    def __init__(self, name=None, pubsub=None, description=None,
                 output_insert_only=None, collapse_updates=None,
                 pulse_interval=None, exp_max_string=None, index_type=None,
                 pubsub_index_type=None, polygon_compute_distance_to=None,
                 proximity_analysis=None, geofencing_algorithm=None,
                 coordinate_type=None, autosize_mesh=None,
                 meshfactor_x=None, meshfactor_y=None,
                 max_meshcells_per_geometry=None,
                 output_multiple_results=None,
                 output_sorted_results=None, log_invalid_geometry=None):
        Window.__init__(self, **get_args(locals()))
        self.polygon_compute_distance_to = polygon_compute_distance_to
        self.proximity_analysis = proximity_analysis
        self.geofencing_algorithm = geofencing_algorithm
        self.coordinate_type = coordinate_type
        self.autosize_mesh = autosize_mesh
        self.meshfactor_x = meshfactor_x
        self.meshfactor_y = meshfactor_y
        self.max_meshcells_per_geometry = max_meshcells_per_geometry
        self.output_multiple_results = output_multiple_results
        self.output_sorted_results = output_sorted_results
        self.log_invalid_geometry = log_invalid_geometry
        self.geometry = Geometry()
        self.position = Position()
        self.output = Output()

    def copy(self, deep=False):
        out = Window.copy(self, deep=deep)
        out.polygon_compute_distance_to = self.polygon_compute_distance_to
        out.proximity_analysis = self.proximity_analysis
        out.geofencing_algorithm = self.geofencing_algorithm
        out.coordinate_type = self.coordinate_type
        out.autosize_mesh = self.autosize_mesh
        out.meshfactor_x = self.meshfactor_x
        out.meshfactor_y = self.meshfactor_y
        out.max_meshcells_per_geometry = self.max_meshcells_per_geometry
        out.output_multiple_results = self.output_multiple_results
        out.output_sorted_results = self.output_sorted_results
        out.log_invalid_geometry = self.log_invalid_geometry
        out.geometry = self.geometry.copy(deep=deep) 
        out.position = self.position.copy(deep=deep) 
        out.output = self.output.copy(deep=deep) 

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`GeofenceWindow`

        '''
        out = super(GeofenceWindow, cls).from_element(data, session=session)

        for item in data.findall('./geofence'):
            for key, value in six.iteritems(item.attrib):
                key = key.replace('-', '_')
                if hasattr(out, key):
                    setattr(out, key, value)

        for item in data.findall('./geometry'):
            out.geometry = Geometry.from_element(item, session=session)

        for item in data.findall('./position'):
            out.position = Position.from_element(item, session=session)

        for item in data.findall('./output'):
            out.output = Output.from_element(item, session=session)

        return out

    from_xml = from_element

    def to_element(self, query=None):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = Window.to_element(self, query=query)
        xml.add_elem(out, 'geofence',
                     attrib=dict(polygon_compute_distance_to=self.polygon_compute_distance_to,
                                 proximity_analysis=self.proximity_analysis,
                                 geofencing_algorithm=self.geofencing_algorithm,
                                 coordinate_type=self.coordinate_type,
                                 autosize_mesh=self.autosize_mesh,
                                 meshfactor_x=self.meshfactor_x,
                                 meshfactor_y=self.meshfactor_y,
                                 max_meshcells_per_geometry=self.max_meshcells_per_geometry,
                                 output_multiple_results=self.output_multiple_results,
                                 output_sorted_results=self.output_sorted_results,
                                 log_invalid_geometry=self.log_invalid_geometry))
        xml.add_elem(out, self.geometry.to_element())
        xml.add_elem(out, self.position.to_element())
        xml.add_elem(out, self.output.to_element())
        connectors_to_end(out)
        return out

    def set_geometry(self, desc_fieldname=None, data_fieldname=None,
                     x_fieldname=None, y_fieldname=None,
                     radius_fieldname=None, radius=None,
                     data_separator=None):
        '''
        Set geometry parameters

        Parameters
        ----------
        desc_fieldname : string, optional
            Specifies the name of the geometry input window’s field that
            contains the polygon coordinates data.
        data_fieldname : string, optional
            Specifies the name of the geometry input window’s field that
            contains the geometry description.
        x_fieldname : string, optional
            Specifies the name of the geometry input window’s field that
            contains the location X or longitude coordinate of the
            circle's center.
        y_fieldname : string, optional
            Specifies the name of the geometry input window’s field that
            contains the location Y or latitude coordinate of the
            circle's center.
        radius_fieldname : string, optional
            Specifies the name of the geometry input window’s field that
            contains the circle radius distance.
        radius : int or float, optional
            Specifies the default circle's radius distance.
            Default: 1000
        data_separator : string, optional
            Specifies the coordinate delimiter character used in the
            geometry data field specified by the property data-fieldname.
            Default: <space>

        '''
        for key, value in six.iteritems(get_args(locals())):
            if value is not None:
                setattr(self.geometry, key, value)

    def set_position(self, x_fieldname=None, y_fieldname=None,
                     lookupdistance_fieldname=None, lookupdistance=None):
        '''
        Set position parameters

        Parameters
        ----------
        x_fieldname : string, optional
           Specifies the name of the position input window’s field that
           contains the position X or longitude coordinate.
        y_fieldname : string, optional
           Specifies the name of the position input window’s field that
           contains the position Y or latitude coordinate.
        lookupdistance_fieldname : string, optional
           Specifies the name of the position input window’s field that
           contains the position lookup distance.
        lookupdistance : int or float, optional
           This distance is in units for Cartesian coordinates and in
           meters for geographic coordinates.
           Default: 0.

        '''
        for key, value in six.iteritems(get_args(locals())):
            if value is not None:
                setattr(self.position, key, value)

    def set_output(self, geotype_fieldname=None, geoid_fieldname=None,
                   geodesc_fieldname=None, geodistance_fieldname=None,
                   eventnumber_fieldname=None):
        '''
        Set output parameters

        Parameters
        ----------
        geotype_fieldname : string, optional
            Specifies the name of the output schema field that receives
            the geometry Type.
        geoid_fieldname : string, optional
            Specifies the name of the output schema field that receives
            the geometry ID.
        geodesc_fieldname : string, optional
            Specifies the name of the output schema field that receives
            the geometry description.
        geodistance_fieldname : string, optional
            Specifies the name of the output schema field that receives
            the distance between the position and the geometry (center
            point for circle geometries and centroid for polygons).
        eventnumber_fieldname : string, optional
            Specifies the name of the output schema additional key field
            that receives the generated event number.

        '''
        for key, value in six.iteritems(get_args(locals())):
            if value is not None:
                setattr(self.output, key, value)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Websocket Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class WebSocketPublisher(Connector):
    '''
    Publish websocket events

    Parameters
    ----------
    url : string
        Specifies the URL for the WebSocket connection. 
    configUrl : string
        Specifies the URL for the connector configuration file.
        This configuration file contains information about the
        transformation steps required to publish events.
    contentType : string
        Specifies XML or JSON as the type of content received
        over the WebSocket connection.
    sslCertificate : string, optional
        Specifies the location of the SSL certificate to use
        when connecting to a secure server.
    sslPassphrase : string, optional
        Specifies the password for the SSL certificate.
    requestHeaders : string, optional
        Specifies a comma-separated list of request headers to
        send to the server. The list must consist of
        name-value pairs in name:value format.
    maxevents : int, optional
         Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`WebSocketPublisher`

    '''
    connector_key = dict(cls='websocket', type='publish')

    property_defs = dict(
        url=prop('url', dtype='string', required=True),
        configUrl=prop('configUrl', dtype='string', required=True),
        contentType=prop('contentType', dtype='string', required=True),
        sslCertificate=prop('sslCertificate', dtype='string'),
        sslPassphrase=prop('sslPassphrase', dtype='string'),
        requestHeaders=prop('requestHeaders', dtype='string'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, url, configUrl, contentType, name=None, is_active=None,
                 sslCertificate=None, sslPassphrase=None, requestHeaders=None,
                 maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'websocket', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['url',
                                                   'configUrl',
                                                   'contentType'],
                                         delete='type')
        return cls(req[0], req[1], req[2], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import copy
import six
import unittest
from . import utils as tm
from ..connectors.base import (get_connector_class, listify, map_properties,
                               Connector, get_subclasses)
from ..connectors.fs import FileSubscriber, FilePublisher, SocketPublisher


class TestUtils(tm.TestCase):

    def test_listify(self):
        self.assertEqual(listify(None), None)
        self.assertEqual(listify(1), [1])
        self.assertEqual(listify('a'), ['a'])
        self.assertEqual(listify([1, 2]), [1, 2])
        self.assertEqual(listify((1, 2)), [1, 2])

    def test_map_properties(self):
        properties = dict(fsname='filename.txt', fstype='csv', snapshot=True, type='subscribe')
        req, prop = map_properties(FileSubscriber, properties, required='fsname',
                                   delete='type')
        self.assertEqual(list(req), ['filename.txt'])
        self.assertEqual(prop, dict(fstype='csv', snapshot=True))

        # Bad parameter
        prop['foo'] = 'bar'
        with self.assertRaises(KeyError):
            map_properties(FileSubscriber, prop, required='fsname', delete='type')

        # No required values
        properties = dict(fsname='filename.txt', fstype='csv', snapshot=True, type='subscribe')
        prop = map_properties(FileSubscriber, properties, delete='type')
        self.assertEqual(prop, dict(fstype='csv', snapshot=True, fsname='filename.txt'))

    def test_get_subclasses(self):
        out = list(get_subclasses(Connector))
        self.assertIn(FileSubscriber, out)
        self.assertIn(FilePublisher, out)


class TestConnector(tm.TestCase):

    def test_get_connector_class(self):
        # Publisher XML
        out = get_connector_class(r'''<connector name='pub' class='fs'>
                <properties>
                  <property name='type'>pub</property>
                  <property name='fstype'>csv</property>
                  <property name='fsname'>input/input1.csv</property>
                  <property name='dateformat'>%Y-%m-%d %H:%M:%S</property>
                </properties>
              </connector>''')
        self.assertTrue(out is FilePublisher)

        # Subscriber XML
        out = get_connector_class(r'''<connector name='sub' class='fs'>
                <properties>
                  <property name='type'>sub</property>
                  <property name='fstype'>csv</property>
                  <property name='fsname'>input/input1.csv</property>
                  <property name='dateformat'>%Y-%m-%d %H:%M:%S</property>
                </properties>
              </connector>''')
        self.assertTrue(out is FileSubscriber)

        # Subscriber XML with type attribute
        out = get_connector_class(r'''<connector name='sub' class='fs' type='sub'>
                <properties>
                  <property name='fstype'>csv</property>
                  <property name='fsname'>input/input1.csv</property>
                  <property name='dateformat'>%Y-%m-%d %H:%M:%S</property>
                </properties>
              </connector>''')
        self.assertTrue(out is FileSubscriber)

        # Publisher name
        out = get_connector_class('fs', type='publish',
                                  properties=dict(fstype='csv', fsname='input1.csv',
                                                  dateformat='%Y-%m-%d %H:%M:%S'))
        self.assertTrue(out is FilePublisher)

        # Subscriber name
        out = get_connector_class('fs', type='subscribe',
                                  properties=dict(fstype='csv', fsname='input1.csv',
                                                  dateformat='%Y-%m-%d %H:%M:%S'))
        self.assertTrue(out is FileSubscriber)

        # Unspecified type
        out = get_connector_class('fs',
                                  properties=dict(fstype='csv', fsname='input1.csv',
                                                  dateformat='%Y-%m-%d %H:%M:%S'))
        self.assertTrue(out is FileSubscriber)

        # Unrecognized type
        out = get_connector_class('foo',
                                  properties=dict(fstype='csv', fsname='input1.csv',
                                                  dateformat='%Y-%m-%d %H:%M:%S'))
        self.assertTrue(out is Connector)

        # Regex property value selector
        out = get_connector_class('fs', type='publish',
                                  properties=dict(fstype='csv', fsname=':2900',
                                                  dateformat='%Y-%m-%d %H:%M:%S'))
        self.assertTrue(out is SocketPublisher)


if __name__ == '__main__':
    tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP IBM WebSphene MQ Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class WebSphereMQSubscriber(Connector):
    '''
    Subscribe to IBM WebSphere MQ events

    Parameters
    ----------
    mqtype : string
        Specifies binary, CSV, JSON, XML, or the name of a string
        field in the subscribed window schema.
    snapshot : boolean, optional
        Specifies whether to send snapshot data.
    mqtopic : string, optional
        Specifies the MQ topic name. Required if mqqueue is
        not configured.
    mqqueue : string, optional
        Specifies the MQ queue name. Required if mqtopic is
        not configured.
    collapse : string, optional
        Enables conversion of UPDATE_BLOCK events to make subscriber
        output publishable. The default value is disabled.
    queuemanager : string, optional
        Specifies the MQ queue manager.
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP
        fields in CSV events. The default behavior is these fields
        are interpreted as an integer number of seconds
        (ESP_DATETIME) or microseconds (ESP_TIMESTAMP) since epoch.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks
        received by a subscriber that were introduced by a
        window retention policy.
    configfilesection : string, optional
        Specifies the name of the section in the connector config
        file to parse for configuration parameters. Specifies
        the value as [configfilesection].
    protofile : string, optional
        Specifies the .proto file that contains the Google
        Protocol Buffers message definition. This definition is
        used to convert event blocks to protobuf messages. When
        you specify this parameter, you must also specify the
        protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message
        in the .proto file that you specified with the protofile
        parameter. Event blocks are converted into this message.
    usecorrelid : boolean, optional
        Copies the value of the correlid field in the event to
        the MQ message correlation ID.
    csvmsgperevent : int, optional
        For CSV, specifies to send one message per event. The
        default is one message per transactional event block or
        else one message per event.
    csvmsgpereventblock : int, optional
        For CSV, specifies to send one message per event block.
        The default is one message per transactional event block
        or else one message per event.

    Returns
    -------
    :class:`WebSphereMQSubscriber`

    '''
    connector_key = dict(cls='mq', type='subscribe')

    property_defs = dict(
        mqtype=prop('mqtype', dtype='string', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        mqtopic=prop('mqtopic', dtype='string'),
        mqqueue=prop('mqqueue', dtype='string'),
        collapse=prop('collapse', dtype='string'),
        queuemanager=prop('queuemanager', dtype='string'),
        dateformat=prop('dateformat', dtype='string'),
        rmretdel=prop('rmretdel', dtype='boolean'),
        configfilesection=prop('configfilesection', dtype='string'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        usecorrelid=prop('usecorrelid', dtype='boolean'),
        csvmsgperevent=prop('csvmsgperevent', dtype='int'),
        csvmsgpereventblock=prop('csvmsgpereventblock', dtype='int')
    )

    def __init__(self, mqtype=None, name=None, is_active=None, snapshot=None,
                 mqtopic=None, mqqueue=None, collapse=None,
                 queuemanager=None, dateformat=None, rmretdel=None,
                 configfilesection=None, protofile=None,
                 protomsg=None, usecorrelid=None, csvmsgperevent=None,
                 csvmsgpereventblock=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'mq', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['mqtype'],
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)


class WebSphereMQPublisher(Connector):
    '''
    Publish IBM WebSphere MQ events

    Parameters
    ----------
    mqtype : string
        Specifies binary, CSV, JSON, XML, or opaquestring. For
        opaquestring, the Source window schema is assumed to
        be "index:int64,message:string".
    mqtopic : string, optional
        Specifies the MQ topic name. Required if mqqueue is
        not configured.
    mqqueue : string, optional
        Specifies the MQ queue name. Required if mqtopic is
        not configured.
    mqsubname : string, optional
        Specifies the MQ subscription name. Required if
        mqtopic is configured.
    blocksize : int, optional
        Specifies the number of events to include in a
        published event block. The default value is 1.
    transactional : string, optional
        Sets the event block type to transactional. The
        default value is normal.
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP
        fields in CSV events. The default behavior is these
        fields are interpreted as an integer number of 
        seconds (ESP_DATETIME) or microseconds (ESP_TIMESTAMP)
        since epoch.
    queuemanager : string, optional
        Specifies the MQ queue manager.
    configfilesection : string, optional
        Specifies the name of the section in the connector config
        file to parse for configuration parameters. Specifies
        the value as [configfilesection].
    ignorecsvparseerrors : boolean, optional
        Specifies that when a field in an input CSV event cannot
        be parsed, the event is dropped, an error is logged,
        and publishing continues.
    protofile : string, optional
        Specifies the .proto file that contains the Google
        Protocol Buffers message definition. This definition
        is used to convert event blocks to protobuf messages.
        When you specify this parameter, you must also specify
        the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message
        in the .proto file that you specified with the protofile
        parameter. Event blocks are converted into this message.
    csvfielddelimiter : string, optional
        Specifies the character delimiter for field data in
        input CSV events. The default delimiter is the , character.
    noautogenfield : boolean, optional
        Specifies that input events are missing the key field
        that is autogenerated by the source window.
    publishwithupsert : boolean, optional
        Builds events with opcode=Upsert instead of Insert.
    addcsvopcode : boolean, optional
        Prepends an opcode and comma to input CSV events. The
        opcode is Insert unless publishwithupsert is enabled.
    addcsvflags : string, optional
        Specifies the event type to insert into input CSV
        events (with a comma). Valid values are "normal"
        and "partialupdate".
    usecorrelid : boolean, optional
        Copies the value of the MQ message correlation ID into
        the correlid field in every Event Stream Processing
        event.
    ignoremqmdformat : string, optional
        Specifies to ignore the value of the Message Descriptor
        Format parameter, and assume the message format is
        compatible with the mqtype parameter setting.
    maxevents : int, optional
        Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`WebSphereMQPublisher`

    '''
    connector_key = dict(cls='mq', type='publish')

    property_defs = dict(
        mqtype=prop('mqtype', dtype='string', required=True),
        mqtopic=prop('mqtopic', dtype='string'),
        mqqueue=prop('mqqueue', dtype='string'),
        mqsubname=prop('mqsubname', dtype='string'),
        blocksize=prop('blocksize', dtype='int'),
        transactional=prop('transactional', dtype='string'),
        dateformat=prop('dateformat', dtype='string'),
        queuemanager=prop('queuemanager', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        ignorecsvparseerrors=prop('ignorecsvparseerrors', dtype='boolean'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        csvfielddelimiter=prop('csvfielddelimiter', dtype='string'),
        noautogenfield=prop('noautogenfield', dtype='boolean'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        addcsvopcode=prop('addcsvopcode', dtype='boolean'),
        addcsvflags=prop('addcsvflags', dtype='string'),
        usecorrelid=prop('usecorrelid', dtype='boolean'),
        ignoremqmdformat=prop('ignoremqmdformat', dtype='boolean'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, mqtype=None, name=None, is_active=None,
                 mqtopic=None, mqqueue=None, mqsubname=None,
                 blocksize=None, transactional=None, dateformat=None,
                 queuemanager=None, configfilesection=None,
                 ignorecsvparseerrors=None, protofile=None,
                 protomsg=None, csvfielddelimiter=None,
                 noautogenfield=None, publishwithupsert=None,
                 addcsvopcode=None, addcsvflags=None,
                 usecorrelid=None, ignoremqmdformat=None,
                 maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'mq', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['mqtype'],
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Solace Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class SolaceSubscriber(Connector):
    '''
    Subscribe to Solace events

    Parameters
    ----------
    solhostport : string
        Specifies the appliance to connect to, in the form 'host:port'
    solvpn : string
        Specifies the appliance message VPN to assign the client to
        which the session connects.
    soltopic : string
        Specifies the Solace destination topic to which to publish
    urlhostport : string
        Specifies the host:port field in the metadata topic subscribed
        to on start-up to field metadata requests.
    numbufferedmsgs : int
        Specifies the maximum number of messages buffered by a standby
        subscriber connector.
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    collapse : string, optional
        Enables conversion of UPDATE_BLOCK events to make subscriber
        output publishable. The default value is disabled.
    hotfailover : boolean, optional
        Enables hot failover mode.
    buspersistence : string, optional
        Sets the Solace message delivery mode to Guaranteed Messaging.
        The default value is Direct Messaging.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks
        received by a subscriber that were introduced by a window
        retention policy.
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol
        Buffers message definition used to convert event blocks to protobuf
        messages. When you specify this parameter, you must also specify
        the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in the
        .proto file that you specified with the protofile parameter.
        Event blocks are converted into this message.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file
        to parse for configuration parameters. Specify the value
        as [configfilesection].
    json : boolean, optional
        Enables transport of event blocks encoded as JSON messages
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields in
        CSV events. The default behavior is these fields are interpreted as
        an integer number of seconds (ESP_DATETIME) or microseconds
        (ESP_TIMESTAMP) since epoch.
    solpasswordencrypted : boolean, optional
        Specifies that solpassword is encrypted

    Returns
    -------
    :class:`SolaceSubscriber`

    '''
    connector_key = dict(cls='sol', type='subscribe')

    property_defs = dict(
        solhostport=prop('solhostport', dtype='string', required=True),
        soluserid=prop('soluserid', dtype='string', required=True),
        solpassword=prop('solpassword', dtype='string', required=True),
        solvpn=prop('solvpn', dtype='string', required=True),
        soltopic=prop('soltopic', dtype='string', required=True),
        urlhostport=prop('urlhostport', dtype='string', required=True),
        numbufferedmsgs=prop('numbufferedmsgs', dtype='int', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        collapse=prop('collapse', dtype='string'),
        hotfailover=prop('hotfailover', dtype='boolean'),
        buspersistence=prop('buspersistence', dtype='string'), 
        rmretdel=prop('rmretdel', dtype='boolean'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        json=prop('json', dtype='boolean'),
        dateformat=prop('dateformat', dtype='string'),
        solpasswordencrypted=prop('solpasswordencrypted', dtype='boolean')
    )

    def __init__(self, solhostport=None, soluserid=None, solpassword=None,
                 solvpn=None, soltopic=None, urlhostport=None,
                 name=None, is_active=None,
                 numbufferedmsgs=None, snapshot=None, collapse=None,
                 hotfailover=None, buspersistence=None, rmretdel=None,
                 protofile=None, protomsg=None, configfilesection=None,
                 json=None, dateformat=None, solpasswordencrypted=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'sol', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['solhostport',
                                                   'soluserid',
                                                   'solpassword',
                                                   'solvpn', 'soltopic',
                                                   'urlhostport',
                                                   'numbufferedmsgs'],
                                         delete='type')
        return cls(req[0], req[1], req[2], req[3], req[4], req[5],
                   name=name, is_active=is_active, **properties)


class SolacePublisher(Connector):
    '''
    Publish events to Solace

    Parameters
    ----------
    solhostport : string
        Specifies the appliance to connect to, in the form “host:port”
    soluserid : string
        Specifies the user name required to authenticate the connector’s
        session with the appliance.
    solpassword : string
        Specifies the password associated with soluserid
    solvpn : string
        Specifies the appliance message VPN to assign the client to which
        the session connects.
    soltopic : string
        Specifies the Solace topic to which to subscribe
    urlhostport : string
        Specifies the host:port field in the metadata topic subscribed
        to on start-up to field metadata requests.
    buspersistence : boolean, optional
        Creates the Guaranteed message flow to bind to the topic endpoint
        provisioned on the appliance that the published Guaranteed messages
        are delivered and spooled to
    buspersistencequeue : string, optional
        Specifies the name of the queue to which the Guaranteed message
        flow binds.
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol Buffers
        message definition used to convert event blocks to protobuf
        messages. When you specify this parameter, you must also specify
        the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in the
        .proto file that you specified with the protofile parameter.
        Event blocks are converted into this message.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file
        to parse for configuration parameters. Specify the value
        as [configfilesection].
    json : boolean, optional
        Enables transport of event blocks encoded as JSON messages
    publishwithupsert : boolean, optional
        Specifies to build with opcode = Upsert instead of opcode = Insert
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields
        in CSV events. The default behavior is these fields are
        interpreted as an integer number of seconds (ESP_DATETIME) or
        microseconds (ESP_TIMESTAMP) since epoch.
    solpasswordencrypted : boolean, optional
        Specifies that solpassword is encrypted
    getmsgfromdestattr : boolean, optional
        Specifies to extract the payload from the destination attribute
        instead of the message body.
    transactional : string, optional
        When getmsgfromdestattr is enabled, sets the event block type
        to transactional. The default value is normal.
    blocksize : int, optional
        When getmsgfromdestattr is enabled, specifies the number of
        events to include in a published event block. The default
        value is 1.
    maxevents : int, optional
        Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`SolacePublisher`

    '''
    connector_key = dict(cls='sol', type='publish')

    property_defs = dict(
        solhostport=prop('solhostport', dtype='string', required=True),
        soluserid=prop('soluserid', dtype='string', required=True),
        solpassword=prop('solpassword', dtype='string', required=True),
        solvpn=prop('solvpn', dtype='string', required=True),
        soltopic=prop('soltopic', dtype='string', required=True),
        urlhostport=prop('urlhostport', dtype='string', required=True),
        buspersistence=prop('buspersistence', dtype='boolean'),
        buspersistencequeue=prop('buspersistencequeue', dtype='string'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        json=prop('json', dtype='boolean'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        dateformat=prop('dateformat', dtype='string'),
        solpasswordencrypted=prop('solpasswordencrypted', dtype='boolean'),
        getmsgfromdestattr=prop('getmsgfromdestattr', dtype='boolean'),
        transactional=prop('transactional', dtype='string'),
        blocksize=prop('blocksize', dtype='int'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, solhostport=None, soluserid=None,
                 solpassword=None, solvpn=None, soltopic=None,
                 urlhostport=None, name=None, is_active=None,
                 buspersistence=None, buspersistencequeue=None,
                 protofile=None, protomsg=None, configfilesection=None,
                 json=None, publishwithupsert=None, dateformat=None,
                 solpasswordencrypted=None, getmsgfromdestattr=None,
                 transactional=None, blocksize=None, maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'sol', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['solhostport',
                                                   'soluserid',
                                                   'solpassword',
                                                   'solvpn',
                                                   'soltopic',
                                                   'urlhostport'],
                                         delete='type')
        return cls(req[0], req[1], req[2], req[3], req[4], req[5],
                   name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

import re
from .base import BaseWindow, attribute
from .features import (FinalizedCallbackFeature, SchemaFeature,
                       FunctionContextFeature)
from .utils import get_args, ensure_element, listify, connectors_to_end
from ..utils import xml


class SMTPSettings(object):
    '''
    SMTP server settings

    Parameters
    ----------
    host : string
        SMTP host name
    port : int, optional
        SMTP port number
    user : string, optional
        User name
    password : string, optional
        Password

    Returns
    -------
    :class:`SMTPSettings`

    '''

    def __init__(self, host, port=None, user=None, password=None):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    def copy(self, deep=False):
        return type(self)(self.host, port=self.port, user=self.user,
                          password=self.password)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`SMTPSettings`

        '''
        data = ensure_element(data)
        out = cls(data.attrib['host'])
        out.user = data.attrib.get('user')
        out.password = data.attrib.get('password')
        out.port = data.attrib.get('port')
        if out.port is not None:
            out.port = int(out.port)
        return out

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('smtp',
                            attrib=dict(host=self.host, port=self.port,
                                        user=self.user, password=self.password))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class EMailMessage(object):
    '''
    EMail notifier

    Parameters
    ----------
    sender : string
        The address of the message sender
    recipients : list-of-strings
        The addresses of the message recipients
    subject : string
        The subject of the message
    from_ : string, optional
        The string to display in the From field of the message
    to : string, optional
        The string to display in the To field of the message
    text : string, optional
        The text of the message
    html : string, optional
        The HTML content of the message
    image : string, optional
        The image content of the message
    deliver : string, optional
        Enclosed text specifies a function to run in order
        to determine whether a notification should be sent.
    name : string, optional
        The name of the notification window
    unresolved_text : string, optional
        Specifies the text to use when the token cannot be resolved.
    throttle_interval : string, optional
        Specifies a time period in which at most one notification
        is sent to a recipient.
    test : boolean, optional
        Specifies whether to run in test mode.

    Returns
    -------
    :class:`EMailMessage`

    '''

    def __init__(self, sender, recipients, subject, from_=None, to=None,
                 text=None, html=None, image=None, deliver=True, name=None,
                 unresolved_text=None, throttle_interval=None, test=False):
        self.sender = sender
        self.recipients = listify(recipients) or []
        self.subject = subject
        self.from_ = from_
        self.to = listify(to) or []
        self.text = listify(text) or []
        self.html = listify(html) or []
        self.image = listify(image) or []
        self.deliver = deliver
        self.name = name
        self.unresolved_text = unresolved_text
        self.throttle_interval = throttle_interval
        self.test = test
        if self.from_ is None:
            self.from_ = sender
        if not self.to:
            self.to = self.recipients

    def copy(self, deep=False):
        return type(self)(self.sender, self.recipients, self.subject, from_=self.from_,
                          to=self.to, text=self.text, html=self.html, image=self.image,
                          deliver=self.deliver, name=self.name,
                          unresolved_text=self.unresolved_text,
                          throttle_interval=self.throttle_interval, test=self.test)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`EMailMessage`

        '''
        data = ensure_element(data)
        out = cls('', '', '')

        out.name = data.attrib.get('name')
        out.unresolved_text = data.attrib.get('unresolved-text')
        out.throttle_interval = data.attrib.get('throttle-interval')
        out.test = data.attrib.get('test', False)

        for item in data.findall('./deliver'):
            out.deliver = item.text

        for item in data.findall('./email-info/sender'):
            out.sender = item.text
        for item in data.findall('./email-info/recipients'):
            out.recipients = re.split(r'\s*,\s*', item.text.strip())
        for item in data.findall('./email-info/subject'):
            out.subject = item.text
        for item in data.findall('./email-info/from'):
            out.from_ = item.text
        for item in data.findall('./email-info/to'):
            out.to = re.split(r'\s*,\s*', item.text.strip())

        for item in data.findall('./email-contents/text-content'):
            out.text.append(item.text)
        for item in data.findall('./email-contents/html-content'):
            out.html.append(item.text)
        for item in data.findall('./email-contents/image-content'):
            out.image.append(item.text)

        return out

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('email', attrib=dict(name=self.name,
                                                unresolved_text=self.unresolved_text,
                                                throttle_interval=self.throttle_interval,
                                                test=self.test))

        xml.add_elem(out, 'deliver', text_content=int(self.deliver))

        info = xml.add_elem(out, 'email-info')
        xml.add_elem(info, 'sender', text_content=self.sender)
        xml.add_elem(info, 'recipients', text_content=','.join(self.recipients))
        xml.add_elem(info, 'subject', text_content=self.subject)
        xml.add_elem(info, 'from', text_content=self.from_)
        xml.add_elem(info, 'to', text_content=', '.join(self.to))

        contents = xml.add_elem(out, 'email-contents')
        for i, item in enumerate(self.text):
            xml.add_elem(contents, 'text-content',
                         attrib=dict(name='text_content_%s' % i),
                         text_content=item)
        for i, item in enumerate(self.html):
            xml.add_elem(contents, 'html-content',
                         attrib=dict(name='html_content_%s' % i),
                         text_content=item)
        for i, item in enumerate(self.image):
            xml.add_elem(contents, 'image-content',
                         attrib=dict(name='image_content_%s' % i),
                         text_content=item)

        return out

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class SMSMessage(object):
    '''
    SMS notifier

    Parameters
    ----------
    sender : string
        The address of the message sender
    subject : string
        The subject of the message
    from_ : string, optional
        The string to display in the From field of the message
    gateway : string, optional
        Specifies the recipient's provider's SMS gateway
    phone : string, optional
        The phone number to send message to
    text : string, optional
        The text of the message
    deliver : string, optional
        Enclosed text specifies a function to run in order
        to determine whether a notification should be sent.
    name : string, optional
        The name of the notification window
    unresolved_text : string, optional
        Specifies the text to use when the token cannot be resolved.
    throttle_interval : string, optional
        Specifies a time period in which at most one notification
        is sent to a recipient.
    test : boolean, optional
        Specifies whether to run in test mode.

    Returns
    -------
    :class:`SMSMessage`

    '''

    def __init__(self, sender, subject, from_=None, gateway=None, phone=None,
                 text=None, deliver=None, name=None, unresolved_text=None,
                 throttle_interval=None, test=None):
        self.sender = sender
        self.subject = subject
        self.from_ = from_
        self.gateway = gateway
        self.phone = phone
        self.text = listify(text) or []
        self.deliver = deliver
        self.name = name
        self.unresolved_text = unresolved_text
        self.throttle_interval = throttle_interval
        self.test = test

    def copy(self, deep=False):
        return type(self)(self.sender, self.subject, from_=self.from_,
                          gateway=self.gateway, phone=self.phone, text=self.text,
                          deliver=self.deliver, name=self.name,
                          unresolved_text=self.unresolved_text,
                          throttle_interval=self.throttle_interval, test=self.test)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`SMSMessage`

        '''
        data = ensure_element(data)
        out = cls('', '')

        out.name = data.attrib.get('name')
        out.unresolved_text = data.attrib.get('unresolved-text')
        out.throttle_interval = data.attrib.get('throttle-interval')
        out.test = data.attrib.get('test', False)

        for item in data.findall('./deliver'):
            out.deliver = item.text

        for item in data.findall('./sms-info/sender'):
            out.sender = item.text
        for item in data.findall('./sms-info/subject'):
            out.subject = item.text
        for item in data.findall('./sms-info/from'):
            out.from_ = item.text
        for item in data.findall('./sms-info/gateway'):
            out.gateway = item.text
        for item in data.findall('./sms-info/phone'):
            out.phone = item.text

        for item in data.findall('./sms-contents/text-content'):
            out.text.append(item.text)

        return out

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('sms', attrib=dict(name=self.name,
                                              unresolved_text=self.unresolved_text,
                                              throttle_interval=self.throttle_interval,
                                              test=self.test))

        xml.add_elem(out, 'deliver', text_content=int(self.deliver))

        info = xml.add_elem(out, 'sms-info')
        xml.add_elem(info, 'sender', text_content=self.sender)
        xml.add_elem(info, 'subject', text_content=self.subject)
        xml.add_elem(info, 'from', text_content=self.from_)
        xml.add_elem(info, 'gateway', text_content=self.gateway)
        xml.add_elem(info, 'phone', text_content=self.phone)

        contents = xml.add_elem(out, 'sms-contents')
        for i, item in enumerate(self.text):
            xml.add_elem(contents, 'text-content',
                         attrib=dict(name='text_content_%s' % i),
                         text_content=item)

        return out

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class MMSMessage(object):
    '''
    Add an SMS notifier

    Parameters
    ----------
    sender : string
        The address of the message sender
    subject : string
        The subject of the message
    from_ : string, optional
        The string to display in the From field of the message
    gateway : string, optional
        Specifies the recipient's provider's SMS gateway
    phone : string, optional
        The phone number to send message to
    text : string, optional
        The text of the message
    image : string, optional
        The image content of the message
    deliver : string, optional
        Enclosed text specifies a function to run in order
        to determine whether a notification should be sent.
    name : string, optional
        The name of the notification window
    unresolved_text : string, optional
        Specifies the text to use when the token cannot be resolved.
    throttle_interval : string, optional
        Specifies a time period in which at most one notification
        is sent to a recipient.
    test : boolean, optional
        Specifies whether to run in test mode.

    '''

    def __init__(self, sender, subject, from_=None, gateway=None, phone=None,
                 text=None, image=None, deliver=None, name=None,
                 unresolved_text=None, throttle_interval=None, test=None):
        self.sender = sender
        self.subject = subject
        self.from_ = from_
        self.gateway = gateway
        self.phone = phone
        self.text = listify(text) or []
        self.image = listify(image) or []
        self.deliver = deliver
        self.name = name
        self.unresolved_text = unresolved_text
        self.throttle_interval = throttle_interval
        self.test = test

    def copy(self, deep=False):
        return type(self)(self.sender, self.subject, from_=self.from_,
                          gateway=self.gateway, phone=self.phone, text=self.text,
                          image=self.image, deliver=self.deliver, name=self.name,
                          unresolved_text=self.unresolved_text,
                          throttle_interval=self.throttle_interval, test=self.test)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`MMSMessage`

        '''
        data = ensure_element(data)
        out = cls('', '')

        out.name = data.attrib.get('name')
        out.unresolved_text = data.attrib.get('unresolved-text')
        out.throttle_interval = data.attrib.get('throttle-interval')
        out.test = data.attrib.get('test', False)

        for item in data.findall('./deliver'):
            out.deliver = item.text

        for item in data.findall('./mms-info/sender'):
            out.sender = item.text
        for item in data.findall('./mms-info/subject'):
            out.subject = item.text
        for item in data.findall('./mms-info/from'):
            out.from_ = item.text
        for item in data.findall('./mms-info/gateway'):
            out.gateway = item.text
        for item in data.findall('./mms-info/phone'):
            out.phone = item.text

        for item in data.findall('./mms-contents/text-content'):
            out.text.append(item.text)
        for item in data.findall('./mms-contents/image-content'):
            out.image.append(item.text)

        return out

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('mms', attrib=dict(name=self.name,
                                              unresolved_text=self.unresolved_text,
                                              throttle_interval=self.throttle_interval,
                                              test=self.test))

        xml.add_elem(out, 'deliver', text_content=int(self.deliver))

        info = xml.add_elem(out, 'mms-info')
        xml.add_elem(info, 'sender', text_content=self.sender)
        xml.add_elem(info, 'subject', text_content=self.subject)
        xml.add_elem(info, 'from', text_content=self.from_)
        xml.add_elem(info, 'gateway', text_content=self.gateway)
        xml.add_elem(info, 'phone', text_content=self.phone)

        contents = xml.add_elem(out, 'mms-contents')
        for i, item in enumerate(self.text):
            xml.add_elem(contents, 'text-content',
                         attrib=dict(name='text_content_%s' % i),
                         text_content=item)
        for i, item in enumerate(self.image):
            xml.add_elem(contents, 'image-content',
                         attrib=dict(name='image_content_%s' % i),
                         text_content=item)

        return out

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class NotificationWindow(BaseWindow, FinalizedCallbackFeature,
                         SchemaFeature, FunctionContextFeature):
    '''
    Notification window

    Parameters
    ----------
    name : string, optional
        The name of the window
    schema : Schema, optional
        The schema of the window

    Attributes
    ----------
    smtp : SMTPSettings
        The SMTP server settings
    email : list-of-EMailMessages
        The email messages to send
    sms : list-of-SMSMessages
        The SMS messages to send
    mms : list-of-MMSMessages
        The MMS messages to send

    Returns
    -------
    :class:`NotificationWindow`

    '''

    window_type = 'notification'

    def __init__(self, name=None, schema=None, pubsub=None, description=None):
        BaseWindow.__init__(self, **get_args(locals()))
        self.smtp = SMTPSettings('localhost')
        self.email = []
        self.sms = []
        self.mms = []

    def copy(self, deep=False):
        out = BaseWindow.copy(self, deep=deep)
        out.smtp = self.smtp.copy(deep=deep)
        if deep:
            out.email = [x.copy(deep=deep) for x in self.email]
            out.sms = [x.copy(deep=deep) for x in self.sms]
            out.mms = [x.copy(deep=deep) for x in self.mms]
        else:
            out.email = list(self.email)
            out.sms = list(self.sms)
            out.mms = list(self.mms)
        return out

    def set_smtp_settings(self, host, port=None, user=None, password=None):
        '''
        Set the SMTP server settings

        Parameters
        ----------
        host : string
            The hostname of the SMTP server
        port : int, optional
            The SMTP server port
        user : string, optional
            The user name on the SMTP server
        password : string, optional
            The password on the SMTP server

        '''
        self.smtp = SMTPSettings(host, port=port, user=user, password=password)

    def add_email(self, sender, recipients, subject, from_=None, to=None,
                  text=None, html=None, image=None, deliver=True, name=None,
                  unresolved_text=None, throttle_interval=None, test=False):
        '''
        Add an email notifier

        Parameters
        ----------
        sender : string
            The address of the message sender
        recipients : list-of-strings
            The addresses of the message recipients
        subject : string
            The subject of the message
        from_ : string, optional
            The string to display in the From field of the message
        to : string, optional
            The string to display in the To field of the message
        text : string, optional
            The text of the message
        html : string, optional
            The HTML content of the message
        image : string, optional
            The image content of the message
        deliver : string, optional
            Enclosed text specifies a function to run in order
            to determine whether a notification should be sent.
        name : string, optional
            The name of the notification window
        unresolved_text : string, optional
            Specifies the text to use when the token cannot be resolved.
        throttle_interval : string, optional
            Specifies a time period in which at most one notification
            is sent to a recipient.
        test : boolean, optional
            Specifies whether to run in test mode.

        '''
        self.email.append(
            EMailMessage(sender, recipients, subject, from_=from_, to=to,
                         text=text, html=html, image=image, deliver=deliver,
                         name=name, unresolved_text=unresolved_text,
                         throttle_interval=throttle_interval, test=test))

    def add_sms(self, sender, subject, from_, gateway, phone, text=None, deliver=True,
                name=None, unresolved_text=None, throttle_interval=None, test=False):
        '''
        Add an SMS notifier

        Parameters
        ----------
        sender : string
            The address of the message sender
        subject : string
            The subject of the message
        from_ : string, optional
            The string to display in the From field of the message
        gateway : string, optional
            Specifies the recipient's provider's SMS gateway
        phone : string, optional
            The phone number to send message to
        text : string, optional
            The text of the message
        deliver : string, optional
            Enclosed text specifies a function to run in order
            to determine whether a notification should be sent.
        name : string, optional
            The name of the notification window
        unresolved_text : string, optional
            Specifies the text to use when the token cannot be resolved.
        throttle_interval : string, optional
            Specifies a time period in which at most one notification
            is sent to a recipient.
        test : boolean, optional
            Specifies whether to run in test mode.

        '''
        self.sms.append(
            SMSMessage(sender, subject, from_=from_, gateway=gateway, phone=phone,
                       text=text, deliver=deliver, name=name,
                       unresolved_text=unresolved_text,
                       throttle_interval=throttle_interval, test=test))

    def add_mms(self, sender, subject, from_, gateway, phone, text=None, image=None,
                deliver=True, name=None, unresolved_text=None, throttle_interval=None,
                test=False):
        '''
        Add an SMS notifier

        Parameters
        ----------
        sender : string
            The address of the message sender
        subject : string
            The subject of the message
        from_ : string, optional
            The string to display in the From field of the message
        gateway : string, optional
            Specifies the recipient's provider's SMS gateway
        phone : string, optional
            The phone number to send message to
        text : string, optional
            The text of the message
        image : string, optional
            The image content of the message
        deliver : string, optional
            Enclosed text specifies a function to run in order
            to determine whether a notification should be sent.
        name : string, optional
            The name of the notification window
        unresolved_text : string, optional
            Specifies the text to use when the token cannot be resolved.
        throttle_interval : string, optional
            Specifies a time period in which at most one notification
            is sent to a recipient.
        test : boolean, optional
            Specifies whether to run in test mode.

        '''
        self.mms.append(
            MMSMessage(sender, subject, from_=from_, gateway=gateway, phone=phone,
                       text=text, image=image, deliver=deliver, name=name,
                       unresolved_text=unresolved_text,
                       throttle_interval=throttle_interval, test=test))

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`NotificationWindow`

        '''
        data = ensure_element(data)
        out = super(NotificationWindow, cls).from_element(data, session=session)

        for item in data.findall('./smtp'):
            out.smtp = SMTPSettings.from_element(item, session=session)

        for item in data.findall('./delivery-channels/email'):
            out.email.append(EMailMessage.from_element(item, session=session))

        for item in data.findall('./delivery-channels/sms'):
            out.sms.append(SMSMessage.from_element(item, session=session))

        for item in data.findall('./delivery-channels/mms'):
            out.mms.append(MMSMessage.from_element(item, session=session))

        return out

    from_xml = from_element

    def to_element(self, query=None):
        '''
        Convert object to Element

        Parameters
        ----------
        query : string, optional
            The name of the continuous query

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = BaseWindow.to_element(self, query=query)

        xml.add_elem(out, self.smtp.to_element())
        channels = xml.add_elem(out, 'delivery-channels')

        for email in self.email:
            xml.add_elem(channels, email.to_element())

        for sms in self.sms:
            xml.add_elem(channels, sms.to_element())

        for mms in self.mms:
            xml.add_elem(channels, mms.to_element())

        connectors_to_end(out)

        return out



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

from .base import BaseWindow, attribute
from .features import ParametersFeature, InputMapFeature, ConnectorsFeature
from .utils import get_args


class TrainWindow(BaseWindow, ParametersFeature, InputMapFeature, ConnectorsFeature):
    '''
    Train window

    Parameters
    ----------
    name : string, optional
        The name of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level
        value of pubsub is manual, true enables publishing and subscribing
        for the window and false disables it.
    description : string, optional
        Description of the window
    algorithm : string, optional
        The name of the algorithm
    input_map : dict, optional
        Input mappings
    **parameters : keyword parameters, optional
        Parameters to the algorithm

    Returns
    -------
    :class:`TrainWindow`

    '''

    window_type = 'train'

    algorithm = attribute('algorithm', dtype='string')

    def __init__(self, name=None, pubsub=None, description=None,
                 algorithm=None, input_map=None, **parameters):
        BaseWindow.__init__(self, **get_args(locals()))
        self.set_parameters(**parameters)
        self.set_inputs(**(input_map or {}))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import copy
import datetime
import os
import six
import esppy
import sys
import unittest
from esppy.config import ESP_ROOT
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestContinuousQuery(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def test_metadata(self):
        metadata = dict(self.query.metadata.items())

        self.query.metadata.clear()

        self.assertEqual(set(self.query.metadata.keys()), set())

        # Set keys
        self.query.metadata['foo'] = 'bar'
        self.query.metadata['one'] = 2

        self.assertEqual(self.query.metadata, {'foo': 'bar', 'one': '2'})
        self.assertEqual(self.query.metadata['foo'], 'bar')
        self.assertEqual(self.query.metadata['one'], '2')

        with self.assertRaises(KeyError):
            self.query.metadata['none']

        # Update keys
        self.query.metadata['foo'] = 'baz'

        self.assertEqual(self.query.metadata, {'foo': 'baz', 'one': '2'})

        # Delete keys
        del self.query.metadata['foo']

        self.assertEqual(self.query.metadata, {'one': '2'})
        self.assertEqual(self.query.metadata['one'], '2')

        with self.assertRaises(KeyError):
            self.query.metadata['foo']

        del self.query.metadata['one']

        self.assertEqual(self.query.metadata, {})
        self.assertEqual(list(self.query.metadata.keys()), [])
        self.assertEqual(list(self.query.metadata.values()), [])

        # Update any original metadata
        self.query.metadata.update(metadata)

        self.assertEqual(self.query.metadata, metadata)

    def test_copy(self):
        # Shallow copy
        cpy = self.query.copy()

        self.assertTrue(cpy is not self.query)
        self.assertEqual(cpy.name, self.query.name)
        self.assertEqual(cpy.project, self.query.project)
        self.assertEqual(set(cpy.windows.keys()), set(self.query.windows.keys()))

        for key in self.query.windows.keys():
            self.assertTrue(cpy.windows[key] is self.query.windows[key]) 

        cpy = copy.copy(self.query)

        self.assertTrue(cpy is not self.query)
        self.assertEqual(cpy.name, self.query.name)
        self.assertEqual(cpy.project, self.query.project)
        self.assertEqual(set(cpy.windows.keys()), set(self.query.windows.keys()))

        for key in self.query.windows.keys():
            self.assertTrue(cpy.windows[key] is self.query.windows[key])

        # Deep copy
        cpy = self.query.copy(deep=True)

        self.assertTrue(cpy is not self.query)
        self.assertEqual(cpy.name, self.query.name)
        self.assertEqual(cpy.project, self.query.project)
        self.assertEqual(set(cpy.windows.keys()), set(self.query.windows.keys()))

        for key in self.query.windows.keys():
            self.assertTrue(cpy.windows[key] is not self.query.windows[key])

        cpy = copy.deepcopy(self.query)

        self.assertTrue(cpy is not self.query)
        self.assertEqual(cpy.name, self.query.name)
        self.assertEqual(cpy.project, self.query.project)
        self.assertEqual(set(cpy.windows.keys()), set(self.query.windows.keys()))

        for key in self.query.windows.keys():
            self.assertTrue(cpy.windows[key] is not self.query.windows[key])

    def test_fullname(self):
        self.assertEqual(self.query.fullname, 'ESPUnitTestProjectSA.contquery')

    def test_metadata(self):
        metadata = dict(self.query.metadata.items())

        self.query.metadata.clear()

        self.assertEqual(set(self.query.metadata.keys()), set())

        # Set keys
        self.query.metadata['foo'] = 'bar'
        self.query.metadata['one'] = 2

        self.assertEqual(self.query.metadata, {'foo': 'bar', 'one': 2})
        self.assertEqual(self.query.metadata['foo'], 'bar')
        self.assertEqual(self.query.metadata['one'], 2)

        with self.assertRaises(KeyError):
            self.query.metadata['none']

        # Update keys
        self.query.metadata['foo'] = 'baz'

        self.assertEqual(self.query.metadata, {'foo': 'baz', 'one': 2})

        # Delete keys
        del self.query.metadata['foo']

        self.assertEqual(self.query.metadata, {'one': 2})
        self.assertEqual(self.query.metadata['one'], 2)

        with self.assertRaises(KeyError):
            self.query.metadata['foo']

        del self.query.metadata['one']

        self.assertEqual(self.query.metadata, {})
        self.assertEqual(list(self.query.metadata.keys()), [])
        self.assertEqual(list(self.query.metadata.values()), [])

        # Update any original metadata
        self.query.metadata.update(metadata)

        self.assertEqual(self.query.metadata, metadata)

    def test_url(self):
        self.assertRegex(self.query.url, r'^https?://.+?/%s/%s/%s/$' %
                                         (ESP_ROOT, self.query.project, self.query.name))

        # Set project to None
        self.query.project = None

        with self.assertRaises(ValueError):
            self.query.url

    def test_to_graph(self):
        out = self.query.to_graph()
        # TODO: Not sure how to verify; just make sure it runs

    def test_repr_svg(self):
        out = self.query._repr_svg_()
        # TODO: Not sure how to verify; just make sure it runs

    def test_str(self):
        qstr = str(self.query)
        self.assertEqual(qstr, "ContinuousQuery(name='%s', project='%s')" % 
                               (self.query.name, self.query.project))

    def test_repr(self):
        qstr = repr(self.query)
        self.assertEqual(qstr, "ContinuousQuery(name='%s', project='%s')" % 
                               (self.query.name, self.query.project))

    def test_from_xml(self):
        xml = self.query.to_xml()
        newquery = self.query.from_xml(xml)

        self.assertEqual(self.query.name, newquery.name)
        self.assertEqual(sorted(self.query.windows.keys()),
                         sorted(newquery.windows.keys()))

    def test_rename_window(self):
        self.assertEqual([x.name for x in self.query['w_data'].targets],
                         ['w_calculate'])
        self.assertEqual([x.name for x in self.query['w_request'].targets],
                         ['w_calculate'])

        self.query.rename_window('w_calculate', 'summary')

        self.assertIn('summary', self.query.windows)
        self.assertNotIn('w_calculate', self.query.windows)

        self.assertEqual([x.name for x in self.query['w_data'].targets],
                         ['summary'])
        self.assertEqual([x.name for x in self.query['w_request'].targets],
                         ['summary'])

    def test_delete_windows(self):
        self.assertEqual(set(self.query.windows.keys()),
                         set(['w_calculate', 'w_data', 'w_request']))

        self.query.delete_windows('w_data', 'w_request')

        self.assertEqual(set(self.query.windows.keys()),
                         set(['w_calculate']))

        self.query.delete_window('w_calculate')

        self.assertEqual(set(self.query.windows.keys()), set())

        with self.assertRaises(KeyError):
            self.query.delete_window('foo')

    def test_add_window(self):
        window = self.s.SourceWindow('new-window')
        self.query.add_window(window)
        self.assertIn('new-window', self.query.windows)
        self.assertEqual(window.name, 'new-window')
        self.assertEqual(window.contquery, self.query.name)
        self.assertEqual(window.project, self.query.project)
        self.assertTrue(window.session is self.query.session)

    def test_get_windows(self):
        # Get all windows
        out = self.query.get_windows()
        self.assertEqual(len(out), 3)
        self.assertEqual(list(sorted(out.keys())), ['contquery.w_calculate',
                                                    'contquery.w_data',
                                                    'contquery.w_request'])

        # Subset names
        out = self.query.get_windows('w_data|w_request')
        self.assertEqual(len(out), 2)
        self.assertEqual(list(sorted(out.keys())), ['contquery.w_data',
                                                    'contquery.w_request'])

        # Non-existent names
        out = self.query.get_windows('foo')
        self.assertEqual(len(out), 0)

        # Filters
        out = self.query.get_windows(filter='match(name,"data|request")')
        self.assertEqual(len(out), 2)
        self.assertEqual(list(sorted(out.keys())), ['contquery.w_data',
                                                    'contquery.w_request'])

    def test_get_window(self):
        # Get specific window
        out = self.query.get_window('w_data')
        self.assertEqual(out.name, 'w_data')
        self.assertTrue(len(out.schema.fields) > 0)

        # Non-existent name
        with self.assertRaises(KeyError):
            self.query.get_window('foo')

    def test_mapping_methods(self):
        from ..windows import SourceWindow

        # __getitem__
        self.assertEqual(self.query['w_data'], self.query.windows['w_data'])
        
        # __setitem__
        window = SourceWindow(name='new_window')
        self.query['new_name'] = window
        self.assertIn('new_name', self.query)
        self.assertEqual(window.name, 'new_name')
        self.assertEqual(window.contquery, self.query.name)
        self.assertEqual(window.project, self.query.project)
        self.assertTrue(window.session, self.query.session)
 
        self.query['w_calculate'].add_target('new_name')
        self.assertEqual([x.name for x in self.query['w_calculate'].targets],
                         ['new_name'])

        # __delitem__
        del self.query['new_name']
        self.assertNotIn('new_name', self.query.windows)
        self.assertNotIn('new_name', self.query)
        self.assertEqual([x.name for x in self.query['w_calculate'].targets], [])

    def test_len(self):
        self.assertEqual(len(self.query.windows), 3)
        self.assertEqual(len(self.query), 3)
        
        del self.query['w_data']

        self.assertEqual(len(self.query.windows), 2)
        self.assertEqual(len(self.query), 2)

    def test_windows_str(self):
        wstr = str(self.query.windows)
        self.assertIn("'w_request': SourceWindow(name='w_request', contquery='contquery', project='ESPUnitTestProjectSA')", wstr)
        self.assertIn("'w_data': SourceWindow(name='w_data', contquery='contquery', project='ESPUnitTestProjectSA')", wstr)
        self.assertIn("'w_calculate': CalculateWindow(name='w_calculate', contquery='contquery', project='ESPUnitTestProjectSA'", wstr)

    def test_windows_repr(self):
        wstr = repr(self.query.windows)
        self.assertIn("'w_request': SourceWindow(name='w_request', contquery='contquery', project='ESPUnitTestProjectSA')", wstr)
        self.assertIn("'w_data': SourceWindow(name='w_data', contquery='contquery', project='ESPUnitTestProjectSA')", wstr)
        self.assertIn("'w_calculate': CalculateWindow(name='w_calculate', contquery='contquery', project='ESPUnitTestProjectSA'", wstr)


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Algorithms '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
import xml.etree.ElementTree as ET
from six.moves import urllib
from .base import ESPObject, attribute
from .utils import xml
from .utils.rest import get_params
from .utils.data import get_project_data


class Algorithm(ESPObject):
    '''
    Base class for all ESP algorithms

    Attributes
    ----------
    name : string
        Name of the algorithm
    parameters : dict
        Dictionary of algorithm parameters
    input_map : dict
        Dictionary of input map entry definitions
    output_map : dict
        Dictionary of output map entry definitions

    Parameters
    ----------
    name : string
        Name of the algorithm

    '''

    reference = attribute('reference', dtype='string')
    algorithm_type = attribute('algorithm-type', dtype='string')
    type = attribute('type', dtype='string')

    def __init__(self, name=None, reference=None, algorithm_type=None, type=None):
        ESPObject.__init__(self)
        self.name = name
        self.algorithm_type = algorithm_type
        self.type = type
        self.reference = reference
        self.parameters = {}
        self.input_map = {}
        self.output_map = {}

    @classmethod
    def from_xml(cls, data, session=None):
        '''
        Create algorithm object from XML definition

        Parameters
        ----------
        data : string
            XML string
        session : requests.Session, optional
            Session the algorithm is associated with

        Returns
        -------
        :class:`Algorithm`

        '''
        if isinstance(data, six.string_types):
            data = xml.from_xml(data) 

        out = cls()
        out.session = session

        out._set_attributes(data.attrib)

        params = out.parameters
        for item in data.findall('./parameters/parameter'):
            attrs = dict(item.attrib)
            name = attrs.pop('name', None)
            params[name] = attrs
            dtype = attrs.get('type', '')
            default = attrs.get('default', None)
            if default is not None:
                if dtype.startswith('int'):
                    params[name]['default'] = int(default)
                elif dtype == 'double':
                    params[name]['default'] = float(default)
                elif dtype == 'boolean':
                    params[name]['default'] = default in ['1', 'true'] and True or False
                else:
                    params[name]['default'] = default.strip()

        input_map = out.input_map
        for item in data.findall('./input-map/input-map-entry'):
            attrs = dict(item.attrib)
            name = attrs.pop('name', None)
            input_map[name] = attrs

        output_map = out.output_map
        for item in data.findall('./output-map/output-map-entry'):
            attrs = dict(item.attrib)
            name = attrs.pop('name', None)
            output_map[name] = attrs

        return out

    def __str__(self):
        attrs = []
        if self.name:
            attrs.append('name=%s' % repr(self.name))
        if self.reference:
            attrs.append('reference=%s' % repr(self.reference))
        if self.type:
            attrs.append('type=%s' % repr(self.type))
        if self.algorithm_type:
            attrs.append('algorithm_type=%s' % repr(self.algorithm_type))
        if self.parameters:
            attrs.append('parameters=%s' % list(sorted(self.parameters)))
        if self.input_map:
            attrs.append('input_map=%s' % list(sorted(self.input_map)))
        if self.output_map:
            attrs.append('output_map=%s' % list(sorted(self.output_map)))
        return '%s(%s)' % (type(self).__name__, ', '.join(attrs))

    def __repr__(self):
        return str(self)
